"""Test functionality to compute the four-component susceptibility tensor for
the Kohn-Sham system."""

from itertools import product, combinations

import numpy as np
import pytest
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.response import ResponseContext, ResponseGroundStateAdapter
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.chi0 import Chi0
from gpaw.response.pair_functions import (get_inverted_pw_mapping,
                                          get_pw_coordinates)
from gpaw.test.conftest import response_band_cutoff

# ---------- chiks parametrization ---------- #


def generate_system_s(spincomponents=['00', '+-']):
    # Compute chiks for different materials and spin components
    system_s = [  # wfs, spincomponent
        ('fancy_si_pw', '00'),
        ('al_pw', '00'),
        ('fe_pw', '00'),
        ('fe_pw', '+-'),
        ('co_pw', '00'),
        ('co_pw', '+-'),
    ]

    # Filter spincomponents
    system_s = [system for system in system_s if system[1] in spincomponents]

    return system_s


def generate_qrel_q():
    # Fractional q-vectors on a path towards a reciprocal lattice vector
    qrel_q = np.array([0., 0.25, 0.5])

    return qrel_q


def get_q_c(wfs, qrel):
    if wfs in ['fancy_si_pw', 'al_pw']:
        # Generate points on the G-X path
        q_c = qrel * np.array([1., 0., 1.])
    elif wfs == 'fe_pw':
        # Generate points on the G-N path
        q_c = qrel * np.array([0., 0., 1.])
    elif wfs == 'co_pw':
        # Generate points on the G-M path
        q_c = qrel * np.array([1., 0., 0.])
    else:
        raise ValueError('Invalid wfs', wfs)

    return q_c


def get_tolerances(system, qrel):
    # Define tolerance for each test system
    wfs, spincomponent = system
    identifier = wfs + '_' + spincomponent

    # Si and Fe the density-density response has perfect symmetry
    atols = {
        'fancy_si_pw_00': 1e-8,
        'fe_pw_00': 1e-8,
    }

    # For the rest, we need to adjust the absolute tolerances. In general
    # it should be possible to lower these tolerances when increasing the
    # number of bands.

    # For Al, the symmetries are not perfectly conserved, but worst for the
    # q-point q_X
    if qrel == 0.0:
        al_atol = 1e-6
    elif qrel == 0.25:
        al_atol = 5e-5
    elif qrel == 0.5:
        al_atol = 2e-4
    atols['al_pw_00'] = al_atol

    # For Fe, the symmetries are not perfectly conserved for the
    # transverse magnetic response
    if qrel == 0.0:
        fet_atol = 3e-3
    elif qrel == 0.25:
        fet_atol = 16e-3
    elif qrel == 0.5:
        fet_atol = 5e-4
    atols['fe_pw_+-'] = fet_atol

    # For the density-density reponse in Co, the symmetries are not perfectly
    # conserved for any of the q-points, but quite well conserved for q = 0
    if qrel == 0.0:
        co_atol = 5e-5
    elif qrel == 0.25:
        co_atol = 5e-3
    elif qrel == 0.5:
        co_atol = 1e-3
    atols['co_pw_00'] = co_atol

    # For the transverse magnetic response in Co, the symmetries are not
    # perfectly conserved for any of the q-points, but again quite well
    # conserved for q = 0
    if qrel == 0.0:
        cot_atol = 5e-4
    elif qrel == 0.25:
        cot_atol = 1e-3
    elif qrel == 0.5:
        cot_atol = 1e-3
    atols['co_pw_+-'] = cot_atol

    if identifier not in atols.keys():
        raise ValueError(system, qrel)

    atol = atols[identifier]
    rtol = 1e-5

    return atol, rtol


def generate_gc_g():
    # Compute chiks both on a gamma-centered and a q-centered pw grid
    gc_g = [True, False]

    return gc_g


def generate_nblocks_n():
    nblocks_n = [1]
    if world.size % 2 == 0:
        nblocks_n.append(2)
    if world.size % 4 == 0:
        nblocks_n.append(4)

    return nblocks_n


# ---------- Actual tests ---------- #


@pytest.mark.response
@pytest.mark.parametrize(
    'system,qrel,gammacentered',
    product(generate_system_s(), generate_qrel_q(), generate_gc_g()))
def test_chiks(in_tmp_dir, gpw_files, system, qrel, gammacentered):
    r"""Test the internals of the ChiKSCalculator.

    In particular, we test that the susceptibility does not change due to the
    details in the internal calculator, such as varrying block distribution,
    band summation scheme, reducing the k-point integral using symmetries or
    basing the ground state adapter on a dynamic (and distributed) GPAW
    calculator.

    Furthermore, we test the symmetries of the calculated susceptibilities.
    """
    # ---------- Inputs ---------- #

    # Part 1: Set up ChiKSTestingFactory
    wfs, spincomponent = system
    atol, rtol = get_tolerances(system, qrel)
    q_c = get_q_c(wfs, qrel)

    ecut = 50
    # Test vanishing and finite real and imaginary frequencies
    frequencies = np.array([0., 0.05, 0.1, 0.2])

    # We add a small (1e-6j) imaginary part to avoid risky floating point
    # operations that may cause NaNs or divide-by-zero.
    complex_frequencies = list(frequencies + 1e-6j) + list(frequencies + 0.1j)
    zd = ComplexFrequencyDescriptor.from_array(complex_frequencies)

    # Part 2: Check toggling of calculation parameters
    # Note: None of these should change the actual results.
    disable_syms_s = [True, False]

    nblocks_n = generate_nblocks_n()
    nn = len(nblocks_n)

    bandsummation_b = ['double', 'pairwise']
    distribution_d = ['GZg', 'ZgG']

    # Symmetry independent tolerances (relating to chiks distribution)
    dist_atol = 1e-8
    dist_rtol = 1e-6

    # Part 3: Check reciprocity and inversion symmetry

    # ---------- Script ---------- #

    # Part 1: Set up ChiKSTestingFactory
    calc = GPAW(gpw_files[wfs], parallel=dict(domain=1))
    nbands = response_band_cutoff[wfs]

    chiks_testing_factory = ChiKSTestingFactory(calc,
                                                spincomponent, q_c, zd,
                                                nbands, ecut, gammacentered)

    # Part 2: Check toggling of calculation parameters

    # Check symmetry toggle and cross-tabulate with nblocks and bandsummation
    chiks_testing_factory.check_parameter_self_consistency(
        parameter='disable_syms', values=disable_syms_s,
        atol=atol, rtol=rtol,
        cross_tabulation=dict(nblocks=nblocks_n,
                              bandsummation=bandsummation_b))

    # Check nblocks and cross-tabulate with disable_syms and bandsummation
    for n1, n2 in combinations(range(nn), 2):
        chiks_testing_factory.check_parameter_self_consistency(
            parameter='nblocks', values=[nblocks_n[n1], nblocks_n[n2]],
            atol=dist_atol, rtol=dist_rtol,
            cross_tabulation=dict(disable_syms=disable_syms_s,
                                  bandsummation=bandsummation_b))

    # Check bandsummation and cross-tabulate with disable_syms and nblocks
    chiks_testing_factory.check_parameter_self_consistency(
        parameter='bandsummation', values=bandsummation_b,
        atol=atol, rtol=rtol,
        cross_tabulation=dict(disable_syms=disable_syms_s,
                              nblocks=nblocks_n))

    # Check internal distribution and cross-tabulate with nblocks
    chiks_testing_factory.check_parameter_self_consistency(
        parameter='distribution', values=distribution_d,
        atol=dist_atol, rtol=dist_rtol,
        cross_tabulation=dict(nblocks=nblocks_n))

    # Part 3: Check reciprocity and inversion symmetry

    # Cross-tabulate disable_syms, nblocks and bandsummation
    chiks_testing_factory.check_reciprocity_and_inversion_symmetry(
        atol=atol, rtol=rtol,
        cross_tabulation=dict(disable_syms=disable_syms_s,
                              nblocks=nblocks_n,
                              bandsummation=bandsummation_b))

    # Cross-tabulate distribution and nblocks
    chiks_testing_factory.check_reciprocity_and_inversion_symmetry(
        atol=atol, rtol=rtol,
        cross_tabulation=dict(distribution=distribution_d,
                              nblocks=nblocks_n))


@pytest.mark.response
@pytest.mark.parametrize(
    'system,qrel',
    product(generate_system_s(spincomponents=['00']), generate_qrel_q()))
def test_chiks_vs_chi0(in_tmp_dir, gpw_files, system, qrel):
    """Test that the ChiKSCalculator is able to reproduce the Chi0Body.

    We use only the default calculation parameter setup for the ChiKSCalculator
    and leave parameter cross-validation to the test above."""

    # ---------- Inputs ---------- #

    # Part 1: chiks calculation
    wfs, spincomponent = system
    atol, rtol = get_tolerances(system, qrel)
    q_c = get_q_c(wfs, qrel)

    ecut = 50
    # Test vanishing and finite real and imaginary frequencies
    frequencies = np.array([0., 0.05, 0.1, 0.2])
    eta = 0.15
    complex_frequencies = frequencies + 1.j * eta

    # Part 2: chi0 calculation

    # Part 3: Check chiks vs. chi0

    # ---------- Script ---------- #

    # Part 1: chiks calculation

    # Initialize ground state adapter
    context = ResponseContext()
    gs = ResponseGroundStateAdapter.from_gpw_file(gpw_files[wfs], context)
    nbands = response_band_cutoff[wfs]

    # Set up complex frequency descriptor
    zd = ComplexFrequencyDescriptor.from_array(complex_frequencies)

    # Calculate chiks
    chiks_calc = ChiKSCalculator(gs, context=context,
                                 ecut=ecut, nbands=nbands)
    chiks = chiks_calc.calculate(spincomponent, q_c, zd)
    chiks = chiks.copy_with_global_frequency_distribution()

    # Part 2: chi0 calculation
    chi0_calc = Chi0(gpw_files[wfs],
                     frequencies=frequencies, eta=eta,
                     ecut=ecut, nbands=nbands,
                     hilbert=False, intraband=False)
    chi0 = chi0_calc.calculate(q_c)
    chi0_wGG = chi0.body.get_distributed_frequencies_array()

    # Part 3: Check chiks vs. chi0
    assert chiks.array == pytest.approx(chi0_wGG, rel=rtol, abs=atol)


# ---------- Test functionality ---------- #


class ChiKSTestingFactory:
    """Factory to calculate and cache chiks objects."""

    def __init__(self, calc,
                 spincomponent, q_c, zd,
                 nbands, ecut, gammacentered):
        self.gs = GSAdapterWithPAWCache(calc)
        self.spincomponent = spincomponent
        self.q_c = q_c
        self.zd = zd
        self.nbands = nbands
        self.ecut = ecut
        self.gammacentered = gammacentered

        self.cached_chiks = {}

    def __call__(self,
                 qsign: int = 1,
                 distribution: str = 'GZg',
                 disable_syms: bool = False,
                 bandsummation: str = 'pairwise',
                 nblocks: int = 1):
        # Compile a string of the calculation parameters for cache look-up
        cache_string = f'{qsign},{distribution},{disable_syms}'
        cache_string += f',{bandsummation},{nblocks}'

        if cache_string in self.cached_chiks:
            return self.cached_chiks[cache_string]

        chiks_calc = ChiKSCalculator(
            self.gs, ecut=self.ecut, nbands=self.nbands,
            gammacentered=self.gammacentered,
            disable_time_reversal=disable_syms,
            disable_point_group=disable_syms,
            bandsummation=bandsummation,
            nblocks=nblocks)

        # Do a manual calculation of chiks
        chiks = chiks_calc._calculate(*chiks_calc._set_up_internals(
            self.spincomponent, qsign * self.q_c, self.zd,
            distribution=distribution))

        chiks = chiks.copy_with_global_frequency_distribution()
        self.cached_chiks[cache_string] = chiks

        return chiks

    def check_parameter_self_consistency(self,
                                         parameter: str, values: list,
                                         atol: float,
                                         rtol: float,
                                         cross_tabulation: dict):
        assert len(values) == 2
        for kwargs in self.generate_cross_tabulated_kwargs(cross_tabulation):
            kwargs[parameter] = values[0]
            chiks1 = self(**kwargs)
            kwargs[parameter] = values[1]
            chiks2 = self(**kwargs)
            compare_pw_bases(chiks1, chiks2)
            assert chiks2.array == pytest.approx(
                chiks1.array, rel=rtol, abs=atol), f'{kwargs}'

    def check_reciprocity_and_inversion_symmetry(self,
                                                 atol: float,
                                                 rtol: float,
                                                 cross_tabulation: dict):
        for kwargs in self.generate_cross_tabulated_kwargs(cross_tabulation):
            # Calculate chiks in q and -q
            chiks1 = self(**kwargs)
            if np.allclose(self.q_c, 0.):
                chiks2 = chiks1
            else:
                chiks2 = self(qsign=-1, **kwargs)
            check_reciprocity_and_inversion_symmetry(chiks1, chiks2,
                                                     atol=atol, rtol=rtol)

    @staticmethod
    def generate_cross_tabulated_kwargs(cross_tabulation: dict):
        # Set up cross tabulation of calculation parameters
        cross_tabulator = product(*[[(key, value)
                                     for value in cross_tabulation[key]]
                                    for key in cross_tabulation])
        for cross_tabulated_parameters in cross_tabulator:
            yield {key: value for key, value in cross_tabulated_parameters}


class GSAdapterWithPAWCache(ResponseGroundStateAdapter):
    """Add a PAW correction cache to the ground state adapter.

    WARNING: Use with care! The cache is only valid, when the plane-wave
    representations are identical.
    """

    def __init__(self, calc):
        super().__init__(calc)

        self._cached_corrections = []
        self._cached_parameters = []

    def pair_density_paw_corrections(self, qpd):
        """Overwrite method with a cached version."""
        cache_index = self._cache_lookup(qpd)
        if cache_index is not None:
            return self._cached_corrections[cache_index]

        return self._calculate_correction(qpd)

    def _calculate_correction(self, qpd):
        correction = super().pair_density_paw_corrections(qpd)

        self._cached_corrections.append(correction)
        self._cached_parameters.append((qpd.q_c, qpd.ecut, qpd.gammacentered))

        return correction

    def _cache_lookup(self, qpd):
        for i, (q_c, ecut,
                gammacentered) in enumerate(self._cached_parameters):
            if np.allclose(qpd.q_c, q_c) and abs(qpd.ecut - ecut) < 1e-8\
               and qpd.gammacentered == gammacentered:
                # Cache hit!
                return i


def compare_pw_bases(chiks1, chiks2):
    """Compare the plane-wave representations of two calculated chiks."""
    G1_Gc = get_pw_coordinates(chiks1.qpd)
    G2_Gc = get_pw_coordinates(chiks2.qpd)
    assert G1_Gc.shape == G2_Gc.shape
    assert np.allclose(G1_Gc - G2_Gc, 0.)


def check_reciprocity_and_inversion_symmetry(chiks1, chiks2, *, atol, rtol):
    """Check the susceptibilities for reciprocity and inversion symmetry

    In particular, we test the reciprocity relation (valid both for μν=00 and
    μν=+-),

    χ_(KS,GG')^(μν)(q, ω) = χ_(KS,-G'-G)^(μν)(-q, ω),

    the inversion symmetry relation,

    χ_(KS,GG')^(μν)(q, ω) = χ_(KS,-G-G')^(μν)(-q, ω),

    and the combination of the two,

    χ_(KS,GG')^(μν)(q, ω) = χ_(KS,G'G)^(μν)(q, ω),

    for a real life periodic systems with an inversion center.

    Unfortunately, there will always be random noise in the wave functions,
    such that these symmetries cannot be fulfilled exactly. Generally speaking,
    the "symmetry" noise can be reduced by running with symmetry='off' in
    the ground state calculation.
    """
    invmap_GG = get_inverted_pw_mapping(chiks1.qpd, chiks2.qpd)

    # Loop over frequencies
    for chi1_GG, chi2_GG in zip(chiks1.array, chiks2.array):
        # Check the reciprocity
        assert chi2_GG[invmap_GG].T == pytest.approx(chi1_GG,
                                                     rel=rtol, abs=atol)
        # Check inversion symmetry
        assert chi2_GG[invmap_GG] == pytest.approx(chi1_GG, rel=rtol, abs=atol)

    # Loop over q-vectors
    for chiks in [chiks1, chiks2]:
        for chiks_GG in chiks.array:  # array = chiks_zGG
            # Check that the full susceptibility matrix is symmetric
            assert chiks_GG.T == pytest.approx(chiks_GG, rel=rtol, abs=atol)



import json
import warnings
from os.path import exists, splitext
import numpy as np
from ase.dft.bandgap import bandgap
from ase.dft.kpoints import get_monkhorst_pack_size_and_offset
from gpaw import GPAW
from gpaw.mpi import rank, serial_comm, world
from gpaw.spinorbit import soc_eigenstates
from gpaw.utilities.blas import gemmdot
from gpaw.ibz2bz import (get_overlap_coefficients,
                         get_phase_shifted_overlap_coefficients)
from gpaw.ibz2bz import get_overlap as get_overlap_new


def get_overlap(calc, bands, u1_nR, u2_nR, P1_ani, P2_ani, dO_aii, bG_v):
    M_nn = np.dot(u1_nR.conj(), u2_nR.T) * calc.wfs.gd.dv
    cell_cv = calc.wfs.gd.cell_cv
    r_av = np.dot(calc.spos_ac, cell_cv)

    for ia in range(len(P1_ani)):
        P1_ni = P1_ani[ia][bands]
        P2_ni = P2_ani[ia][bands]
        phase = np.exp(-1.0j * np.dot(bG_v, r_av[ia]))
        dO_ii = dO_aii[ia]
        M_nn += P1_ni.conj().dot(dO_ii).dot(P2_ni.T) * phase

    return M_nn


def get_berry_phases(calc, spin=0, dir=0, check2d=False):
    if isinstance(calc, str):
        calc = GPAW(calc, communicator=serial_comm, txt=None)
        
    assert len(calc.symmetry.op_scc) == 1  # does not work with symmetry
    gap = bandgap(calc)[0]
    assert gap != 0.0

    M = np.round(calc.get_magnetic_moment())
    assert np.allclose(M, calc.get_magnetic_moment(), atol=0.05), \
        print(M, calc.get_magnetic_moment())
    nvalence = calc.wfs.setups.nvalence
    nocc_s = [int((nvalence + M) / 2), int((nvalence - M) / 2)]
    assert np.allclose(np.sum(nocc_s), nvalence)
    nocc = nocc_s[spin]

    bands = list(range(nocc))
    kpts_kc = calc.get_bz_k_points()
    size = get_monkhorst_pack_size_and_offset(kpts_kc)[0]
    Nk = len(kpts_kc)
    wfs = calc.wfs

    dO_aii = get_overlap_coefficients(wfs)
    
    kd = calc.wfs.kd

    u_knR = []
    proj_k = []
    for k in range(Nk):
        ik = kd.bz2ibz_k[k]
        k_c = kd.bzk_kc[k]
        ik_c = kd.ibzk_kc[ik]
        # Since symmetry is off this should always hold
        assert np.allclose(k_c, ik_c)
        kpt = wfs.kpt_qs[ik][spin]
        psit_nG = kpt.psit_nG
        ut_nR = wfs.gd.empty(nocc, wfs.dtype)

        # Check that all states are occupied
        assert np.all(kpt.f_n[:nocc] > 1e-6)
        N_c = wfs.gd.N_c

        for n in range(nocc):
            ut_nR[n, :] = wfs.pd.ifft(psit_nG[n], ik)
        u_knR.append(ut_nR)

        proj_k.append(kpt.projections)

    indices_kkk = np.arange(Nk).reshape(size)
    tmp = np.concatenate([[i for i in range(3) if i != dir], [dir]])
    indices_kk = indices_kkk.transpose(tmp).reshape(-1, size[dir])

    nkperp = len(indices_kk)
    phases = []
    if check2d:
        phases2d = []
    for indices_k in indices_kk:
        M_knn = []
        for j in range(size[dir]):
            k1 = indices_k[j]
            G_c = np.array([0, 0, 0])
            if j + 1 < size[dir]:
                k2 = indices_k[j + 1]
            else:
                k2 = indices_k[0]
                G_c[dir] = 1
            u1_nR = u_knR[k1]
            u2_nR = u_knR[k2]
            k1_c = kpts_kc[k1]
            k2_c = kpts_kc[k2] + G_c

            if np.any(G_c):
                emiGr_R = np.exp(-2j * np.pi *
                                 np.dot(np.indices(N_c).T, G_c / N_c).T)
                u2_nR = u2_nR * emiGr_R

            bG_c = k2_c - k1_c

            phase_shifted_dO_aii = get_phase_shifted_overlap_coefficients(
                dO_aii, calc.spos_ac, -bG_c)
            M_nn = get_overlap_new(bands,
                                   wfs.gd,
                                   u1_nR,
                                   u2_nR,
                                   proj_k[k1],
                                   proj_k[k2],
                                   phase_shifted_dO_aii)
            M_knn.append(M_nn)
        det = np.linalg.det(M_knn)
        phases.append(np.imag(np.log(np.prod(det))))
        if check2d:
            # In the case of 2D systems we can check the
            # result
            k1 = indices_k[0]
            k1_c = kpts_kc[k1]
            G_c = [0, 0, 1]
            u1_nR = u_knR[k1]
            emiGr_R = np.exp(-2j * np.pi *
                             np.dot(np.indices(N_c).T, G_c / N_c).T)
            u2_nR = u1_nR * emiGr_R

            phase_shifted_dO_aii = get_phase_shifted_overlap_coefficients(
                dO_aii, calc.spos_ac, -bG_c)
            M_nn = get_overlap_new(bands,
                                   calc.wfs.gd,
                                   u1_nR,
                                   u2_nR,
                                   proj_k[k1],
                                   proj_k[k1],
                                   phase_shifted_dO_aii)

            phase2d = np.imag(np.log(np.linalg.det(M_nn)))
            phases2d.append(phase2d)

    # Make sure the phases are continuous
    for p in range(nkperp - 1):
        delta = phases[p] - phases[p + 1]
        phases[p + 1] += np.round(delta / (2 * np.pi)) * 2 * np.pi

    phase = np.sum(phases) / nkperp
    if check2d:
        for p in range(nkperp - 1):
            delta = phases2d[p] - phases2d[p + 1]
            phases2d[p + 1] += np.round(delta / (2 * np.pi)) * 2 * np.pi

        phase2d = np.sum(phases2d) / nkperp

        diff = abs(phase - phase2d)
        if diff > 0.01:
            msg = 'Warning wrong phase: phase={}, 2dphase={}'
            print(msg.format(phase, phase2d))

    return indices_kk, phases


def get_polarization_phase(calc):
    assert isinstance(calc, str)
    name = splitext(calc)[0]
    berryname = '{}-berryphases.json'.format(name)

    phase_c = np.zeros((3,), float)
    if not exists(berryname) and world.rank == 0:
        # Calculate and save berry phases
        calc = GPAW(calc, communicator=serial_comm, txt=None)
        assert len(calc.symmetry.op_scc) == 1
        nspins = calc.wfs.nspins
        data = {}
        for c in [0, 1, 2]:
            data[c] = {}
            for spin in range(nspins):
                indices_kk, phases = get_berry_phases(calc, dir=c, spin=spin)
                data[c][spin] = phases

        # Ionic contribution
        Z_a = []
        for num in calc.atoms.get_atomic_numbers():
            for ida, setup in zip(calc.wfs.setups.id_a,
                                  calc.wfs.setups):
                if abs(ida[0] - num) < 1e-5:
                    break
            Z_a.append(setup.Nv)
        data['Z_a'] = Z_a
        data['spos_ac'] = calc.spos_ac.tolist()

        with open(berryname, 'w') as fd:
            json.dump(data, fd, indent=True)

    world.barrier()
    # Read data and calculate phase
    if world.rank == 0:
        print('Reading berryphases {}'.format(berryname))
    with open(berryname) as fd:
        data = json.load(fd)

    for c in [0, 1, 2]:
        nspins = len(data[str(c)])
        for spin in range(nspins):
            phases = data[str(c)][str(spin)]
            phase_c[c] += np.sum(phases) / len(phases)
    phase_c = phase_c * 2 / nspins

    print(phase_c)
    Z_a = data['Z_a']
    spos_ac = data['spos_ac']
    phase_c += 2 * np.pi * np.dot(Z_a, spos_ac)

    return phase_c


def parallel_transport(calc,
                       direction=0,
                       name=None,
                       scale=1.0,
                       bands=None,
                       theta=0.0,
                       phi=0.0):
    """
    Parallel transport.
    The parallel transport algorithm corresponds to the construction
    of hybrid Wannier functions localized along the Nloc direction.
    While these are not constructed explicitly one may obtain the
    Wannier Charge centers which are given by the eigenvalues of
    the Berry phase matrix (except for a factor of 2*pi) phi_km.
    In addition, one may evaluate the expectation value of spin
    on each of these states along the easy axis (z-axis for
    nonmagnetic systems), which is given by S_km.
    
    Output:
    phi_km, S_km (see above)
    """
    if isinstance(calc, str):
        calc = GPAW(calc, txt=None, communicator=serial_comm)

    if bands is None:
        nv = int(calc.get_number_of_electrons())
        bands = range(nv)

    cell_cv = calc.wfs.gd.cell_cv
    icell_cv = (2 * np.pi) * np.linalg.inv(cell_cv).T
    r_g = calc.wfs.gd.get_grid_point_coordinates()
    Ng = np.prod(np.shape(r_g)[1:]) * 2

    dO_aii = []
    for ia in calc.wfs.kpt_u[0].P_ani.keys():
        dO_ii = calc.wfs.setups[ia].dO_ii
        # Spinor projections require doubling of the (identical) orbitals
        dO_jj = np.zeros((2 * len(dO_ii), 2 * len(dO_ii)), complex)
        dO_jj[::2, ::2] = dO_ii
        dO_jj[1::2, 1::2] = dO_ii
        dO_aii.append(dO_jj)

    N_c = calc.wfs.kd.N_c
    assert 1 in np.delete(N_c, direction)
    Nkx = N_c[0]
    Nky = N_c[1]
    Nkz = N_c[2]

    Nk = Nkx * Nky * Nkz
    Nloc = N_c[direction]
    Npar = Nk // Nloc

    # Parallelization stuff
    myKsize = -(-Npar // (world.size))
    myKrange = range(rank * myKsize,
                     min((rank + 1) * myKsize, Npar))
    myKsize = len(myKrange)

    # Get array of k-point indices of the path. q index is loc direction
    kpts_kq = []
    for k in range(Npar):
        if direction == 0:
            kpts_kq.append(list(range(k, Nkx * Nky, Nky)))
        if direction == 1:
            if Nkz == 1:
                kpts_kq.append(list(range(k * Nky, (k + 1) * Nky)))
            else:
                kpts_kq.append(list(range(k, Nkz * Nky, Nkz)))
        if direction == 2:
            kpts_kq.append(list(range(k * Nloc, (k + 1) * Nloc)))

    G_c = np.array([0, 0, 0])
    G_c[direction] = 1
    G_v = np.dot(G_c, icell_cv)

    kpts_kc = calc.get_bz_k_points()

    if Nloc > 1:
        b_c = kpts_kc[kpts_kq[0][1]] - kpts_kc[kpts_kq[0][0]]
        b_v = np.dot(b_c, icell_cv)
    else:
        b_v = G_v

    soc_kpts = soc_eigenstates(calc,
                               scale=scale,
                               theta=theta,
                               phi=phi)

    def projections(bz_index):
        return soc_kpts[bz_index].P_amj

    def wavefunctions(bz_index):
        return soc_kpts[bz_index].wavefunctions(
            calc, periodic=True)[bands]

    phi_km = np.zeros((Npar, len(bands)), float)
    S_km = np.zeros((Npar, len(bands)), float)
    # Loop over the direction parallel components
    for k in myKrange:
        U_qmm = [np.eye(len(bands))]
        qpts_q = kpts_kq[k]
        # Loop over kpoints in the phase direction
        for q in range(Nloc - 1):
            iq1 = qpts_q[q]
            iq2 = qpts_q[q + 1]
            # print(kpts_kc[iq1], kpts_kc[iq2])
            if q == 0:
                u1_nsG = wavefunctions(iq1)
                P1_ani = projections(iq1)

            u2_nsG = wavefunctions(iq2)
            P2_ani = projections(iq2)

            M_mm = get_overlap(calc,
                               bands,
                               np.reshape(u1_nsG, (len(u1_nsG), Ng)),
                               np.reshape(u2_nsG, (len(u2_nsG), Ng)),
                               P1_ani,
                               P2_ani,
                               dO_aii,
                               b_v)
            V_mm, sing_m, W_mm = np.linalg.svd(M_mm)
            U_mm = np.dot(V_mm, W_mm).conj()
            u_nysxz = np.dot(U_mm, np.swapaxes(u2_nsG, 0, 3))
            u_nxsyz = np.swapaxes(u_nysxz, 1, 3)
            u_nsxyz = np.swapaxes(u_nxsyz, 1, 2)
            u2_nsG = u_nsxyz
            for a in range(len(calc.atoms)):
                P2_ni = P2_ani[a][bands]
                P2_ni = np.dot(U_mm, P2_ni)
                P2_ani[a][bands] = P2_ni
            U_qmm.append(U_mm)
            u1_nsG = u2_nsG
            P1_ani = P2_ani
        U_qmm = np.array(U_qmm)

        # Fix phases for last point
        iq0 = qpts_q[0]
        if Nloc == 1:
            u1_nsG = wavefunctions(iq0)
            P1_ani = projections(iq0)
        u2_nsG = wavefunctions(iq0)
        u2_nsG[:] *= np.exp(-1.0j * gemmdot(G_v, r_g, beta=0.0))
        P2_ani = projections(iq0)

        M_mm = get_overlap(calc,
                           bands,
                           np.reshape(u1_nsG, (len(u1_nsG), Ng)),
                           np.reshape(u2_nsG, (len(u2_nsG), Ng)),
                           P1_ani,
                           P2_ani,
                           dO_aii,
                           b_v)
        V_mm, sing_m, W_mm = np.linalg.svd(M_mm)
        U_mm = np.dot(V_mm, W_mm).conj()
        u_nysxz = np.dot(U_mm, np.swapaxes(u2_nsG, 0, 3))
        u_nxsyz = np.swapaxes(u_nysxz, 1, 3)
        u_nsxyz = np.swapaxes(u_nxsyz, 1, 2)
        u2_nsG = u_nsxyz
        for a in range(len(calc.atoms)):
            P2_ni = P2_ani[a][bands]
            P2_ni = np.dot(U_mm, P2_ni)
            P2_ani[a][bands] = P2_ni

        # Get overlap between first kpts and its smoothly translated image
        u2_nsG[:] *= np.exp(1.0j * gemmdot(G_v, r_g, beta=0.0))
        for a in range(len(calc.atoms)):
            P2_ni = P2_ani[a][bands]
            # P2_ni *= np.exp(1.0j * np.dot(G_v, r_av[a]))
            P2_ani[a][bands] = P2_ni
        u1_nsG = wavefunctions(iq0)
        P1_ani = projections(iq0)
        M_mm = get_overlap(calc,
                           bands,
                           np.reshape(u1_nsG, (len(u1_nsG), Ng)),
                           np.reshape(u2_nsG, (len(u2_nsG), Ng)),
                           P1_ani,
                           P2_ani,
                           dO_aii,
                           np.array([0.0, 0.0, 0.0]))
        l_m, l_mm = np.linalg.eig(M_mm)
        phi_km[k] = np.angle(l_m)

        A_mm = np.zeros_like(l_mm, complex)
        for q in range(Nloc):
            iq = qpts_q[q]
            U_mm = U_qmm[q]
            v_mn = soc_kpts[iq].v_mn
            v_nm = np.einsum('xm, mn -> nx', U_mm, v_mn[bands])
            A_mm += np.dot(v_nm[::2].T.conj(), v_nm[::2])
            A_mm -= np.dot(v_nm[1::2].T.conj(), v_nm[1::2])
        A_mm /= Nloc
        S_km[k] = np.diag(l_mm.T.conj().dot(A_mm).dot(l_mm)).real

    world.sum(phi_km)
    world.sum(S_km)

    if not calc.density.collinear:
        warnings.warn('WARNING: Spin projections are not meaningful ' +
                      'for non-collinear calculations')

    if name is not None:
        np.savez(f'phases_{name}.npz', phi_km=phi_km, S_km=S_km)

    return phi_km, S_km



"""Module defining  ``Eigensolver`` classes."""

from gpaw.eigensolvers.rmmdiis import RMMDIIS
from gpaw.eigensolvers.cg import CG
from gpaw.eigensolvers.davidson import Davidson
from gpaw.eigensolvers.direct import DirectPW
from gpaw.lcao.eigensolver import DirectLCAO
from gpaw.directmin.etdm import ETDM


def get_eigensolver(eigensolver, mode, convergence=None):
    """Create eigensolver object."""
    if eigensolver is None:
        if mode.name == 'lcao':
            eigensolver = 'lcao'
        else:
            eigensolver = 'dav'

    if isinstance(eigensolver, str):
        eigensolver = {'name': eigensolver}

    if isinstance(eigensolver, dict):
        eigensolver = eigensolver.copy()
        name = eigensolver.pop('name')
        eigensolver = {'rmm-diis': RMMDIIS,
                       'cg': CG,
                       'dav': Davidson,
                       'lcao': DirectLCAO,
                       'direct': DirectPW,
                       'etdm': ETDM,
                       }[name](**eigensolver)

    if isinstance(eigensolver, CG):
        eigensolver.tolerance = convergence.get('eigenstates', 4.0e-8)

    assert isinstance(eigensolver, DirectLCAO) == (mode.name == 'lcao') or \
           isinstance(eigensolver, ETDM) == (mode.name == 'lcao')

    return eigensolver



import numpy as np


def CG(A, X, B, maxiter=20, tolerance=1.0e-10, verbose=False):
    """Solve X*A=B using conjugate gradient method.

    ``X`` and ``B`` are ``ndarrays```of shape ``(m, nx, ny, nz)``
    coresponding to matrices of size ``m*n`` (``n=nx*ny*nz``) and
    ``A`` is a callable representing an ``n*n`` matrix::

      A(X, Y)

    will store ``X*A`` in the output array ``Y``.

    On return ``X`` will be the solution to ``X*A=B`` within
    ``tolerance``."""

    m = len(X)
    shape = (m, 1, 1, 1)
    R = np.empty(X.shape, X.dtype.char)
    Q = np.empty(X.shape, X.dtype.char)
    A(X, R)
    R -= B
    P = R.copy()
    c1 = A.sum(np.reshape([abs(np.vdot(r, r)) for r in R], shape))
    for i in range(maxiter):
        error = sum(c1.ravel())
        if verbose:
            print('CG-%d: %e' % (i, error))
        if error < tolerance:
            return i, error
        A(P, Q)
        alpha = c1 / A.sum(np.reshape([np.vdot(q, p)
                                       for p, q in zip(P, Q)], shape))
        X -= alpha * P
        R -= alpha * Q
        c0 = c1
        c1 = A.sum(np.reshape([abs(np.vdot(r, r)) for r in R], shape))
        beta = c1 / c0
        P *= beta
        P += R

    raise ArithmeticError('Did not converge!')



import pytest
from ase.build import molecule
from gpaw import GPAW
import gpaw.dscf as dscf
from gpaw.test import equal

# Ground state calculation


@pytest.mark.dscf
def test_dscf_dscf_lcao():
    calc = GPAW(mode='lcao',
                basis='dzp',
                nbands=8,
                h=0.2,
                xc='PBE',
                spinpol=True,
                convergence={'energy': 100,
                             'density': 1e-3,
                             'bands': -1})

    CO = molecule('CO')
    CO.center(vacuum=3)
    CO.calc = calc

    E_gs = CO.get_potential_energy()

    # Excited state calculation

    calc_es = GPAW(mode='lcao',
                   basis='dzp',
                   nbands=8,
                   h=0.2,
                   symmetry='off',
                   xc='PBE',
                   spinpol=True,
                   convergence={'energy': 100,
                                'density': 1e-3,
                                'bands': -1})

    CO.calc = calc_es
    lumo = dscf.MolecularOrbital(calc,
                                 weights={0: [0, 0, 0, 1], 1: [0, 0, 0, -1]})
    dscf.dscf_calculation(calc_es, [[1.0, lumo, 1]], CO)

    E_es = CO.get_potential_energy()
    dE = E_es - E_gs
    print(dE)
    equal(dE, 5.7595110076, 0.011)



"""Point-group object."""
from typing import Dict
import numpy as np
from gpaw.typing import Array2D


class PointGroup:
    def __init__(self, name: str):
        """Point-group object.

        Name must be one of: C2, C2v, C3v, D2d, D3h, D5, D5h,
        Ico, Ih, Oh, Td or Th.
        """
        import gpaw.point_groups.groups as groups
        self.name = name
        group = getattr(groups, name)()
        self.character_table = np.array(group.character_table)
        self.operations: Dict[str, Array2D] = {}
        for opname, op in group.operations:
            assert opname not in self.operations, opname
            if not isinstance(op, np.ndarray):
                op = op(np.eye(3))
            self.operations[opname] = op
        self.symmetries = group.symmetries
        self.nops = group.nof_operations
        self.complex = getattr(group, 'complex', False)
        self.translations = [self.symmetries[t]
                             for t in (group.Tx_i, group.Ty_i, group.Tz_i)]

    def __str__(self) -> str:
        lines = [[self.name] + list(self.operations)]
        for sym, coefs in zip(self.symmetries, self.character_table):
            lines.append([sym] + list(coefs))
        return '\n'.join(f'{line[0]:5}' +
                         ''.join(f'{word:>10}' for word in line[1:])
                         for line in lines) + '\n'

    def get_normalized_table(self) -> Array2D:
        """Normalized character table."""
        # Divide by degeneracies:
        return self.character_table / self.character_table[:, :1]



import pytest
from ase.lattice.cubic import BodyCenteredCubic

from gpaw import GPAW, PW
from gpaw.test import equal


@pytest.mark.later
def test_xc_qna_spinpol(in_tmp_dir):
    QNA = {'alpha': 2.0,
           'name': 'QNA',
           'orbital_dependent': False,
           'parameters': {'Fe': (0.1485, 0.005)},
           'setup_name': 'PBE',
           'type': 'qna-gga'}

    atoms = BodyCenteredCubic(symbol='Fe',
                              latticeconstant=2.854,
                              pbc=(1, 1, 1))

    atoms.set_initial_magnetic_moments([2, 2])

    calc = GPAW(mode=PW(400),
                kpts=(3, 3, 3),
                experimental={'niter_fixdensity': 2},
                xc=QNA,
                parallel={'domain': 1},
                txt='qna_spinpol.txt')

    atoms.calc = calc
    atoms.get_potential_energy()
    magmoms = atoms.get_magnetic_moments()

    tol = 0.003
    equal(2.252, magmoms[0], tol)
    equal(2.252, magmoms[1], tol)



import json
from os.path import exists, splitext, isfile
from os import remove
from glob import glob

import numpy as np
from gpaw import GPAW
from gpaw.mpi import world
from gpaw.berryphase import get_polarization_phase
from ase.parallel import paropen
from ase.units import Bohr


def get_wavefunctions(atoms, name, params):
    params['symmetry'] = {'point_group': False,
                          'time_reversal': False}
    tmp = splitext(name)[0]
    atoms.calc = GPAW(txt=tmp + '.txt', **params)
    atoms.get_potential_energy()
    atoms.calc.write(name, 'all')
    return atoms.calc


def borncharges(calc, delta=0.01):

    params = calc.parameters
    atoms = calc.atoms
    cell_cv = atoms.get_cell() / Bohr
    vol = abs(np.linalg.det(cell_cv))
    sym_a = atoms.get_chemical_symbols()

    Z_a = []
    for num in calc.atoms.get_atomic_numbers():
        for ida, setup in zip(calc.wfs.setups.id_a,
                              calc.wfs.setups):
            if abs(ida[0] - num) < 1e-5:
                break
        Z_a.append(setup.Nv)
    Z_a = np.array(Z_a)

    # List for atomic indices
    indices = list(range(len(sym_a)))

    pos_av = atoms.get_positions()
    avg_v = np.sum(pos_av, axis=0) / len(pos_av)
    pos_av -= avg_v
    atoms.set_positions(pos_av)
    Z_avv = []
    norm_c = np.linalg.norm(cell_cv, axis=1)
    proj_cv = cell_cv / norm_c[:, np.newaxis]

    B_cv = np.linalg.inv(cell_cv).T * 2 * np.pi
    area_c = np.zeros((3,), float)
    area_c[[2, 1, 0]] = [np.linalg.norm(np.cross(B_cv[i], B_cv[j]))
                         for i in range(3)
                         for j in range(3) if i < j]

    if world.rank == 0:
        print('Atomnum Atom Direction Displacement')
    for a in indices:
        phase_scv = np.zeros((2, 3, 3), float)
        for v in range(3):
            for s, sign in enumerate([-1, 1]):
                if world.rank == 0:
                    print(sym_a[a], a, v, s)
                # Update atomic positions
                atoms.positions = pos_av
                atoms.positions[a, v] = pos_av[a, v] + sign * delta
                prefix = 'born-{}-{}{}{}'.format(delta, a,
                                                 'xyz'[v],
                                                 ' +-'[sign])
                name = prefix + '.gpw'
                berryname = prefix + '-berryphases.json'
                if not exists(name) and not exists(berryname):
                    calc = get_wavefunctions(atoms, name, params)

                try:
                    phase_c = get_polarization_phase(name)
                except ValueError:
                    calc = get_wavefunctions(atoms, name, params)
                    phase_c = get_polarization_phase(name)

                phase_scv[s, :, v] = phase_c

                if exists(berryname):  # Calculation done?
                    if world.rank == 0:
                        # Remove gpw file
                        if isfile(name):
                            remove(name)

        dphase_cv = (phase_scv[1] - phase_scv[0])
        dphase_cv -= np.round(dphase_cv / (2 * np.pi)) * 2 * np.pi
        dP_cv = (area_c[:, np.newaxis] / (2 * np.pi)**3 *
                 dphase_cv)
        dP_vv = np.dot(proj_cv.T, dP_cv)
        Z_vv = dP_vv * vol / (2 * delta / Bohr)
        Z_avv.append(Z_vv)

    data = {'Z_avv': [Z_aa.tolist() for Z_aa in Z_avv],
            'indices_a': indices,
            'sym_a': sym_a}

    filename = 'borncharges-{}.json'.format(delta)

    with paropen(filename, 'w') as fd:
        json.dump(data, fd)

    world.barrier()
    if world.rank == 0:
        files = glob('born-*.gpw')
        for f in files:
            if isfile(f):
                remove(f)



from unittest import SkipTest

from ase import Atoms

from gpaw import GPAW
from gpaw.mpi import size


def test_noncollinear_o2(in_tmp_dir):
    if size > 2:
        raise SkipTest()

    a = Atoms('OO', [[0, 0, 0], [0, 0, 1.1]], magmoms=[1, 1], pbc=(1, 0, 0))
    a.center(vacuum=2.5)
    a.calc = GPAW(mode='pw',
                  kpts=(2, 1, 1))
    f0 = a.get_forces()

    a.calc = GPAW(mode='pw',
                  kpts=(2, 1, 1),
                  symmetry='off',
                  experimental={'magmoms': [[0, 0.5, 0.5], [0, 0, 1]]})
    f = a.get_forces()

    assert abs(f - f0).max() < 0.01

    a.calc.write('o2.gpw')
    a.calc.write('o2w.gpw', 'all')
    GPAW('o2w.gpw')



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Grid transformers.

This module defines tools for doing interpolations/restrictions between
differentt uniform 3D grids.
"""

import numpy as np

from gpaw import debug
from gpaw.utilities import is_contiguous
import _gpaw


class _Transformer:
    def __init__(self, gdin, gdout, nn=1, dtype=float, xp=np):
        self.gdin = gdin
        self.gdout = gdout
        self.nn = nn
        assert 1 <= nn <= 4
        self.dtype = dtype
        self.xp = xp

        pad_cd = np.empty((3, 2), int)
        neighborpad_cd = np.empty((3, 2), int)
        skip_cd = np.empty((3, 2), int)

        if (gdin.N_c == 2 * gdout.N_c).all():
            # Restriction:
            pad_cd[:, 0] = 2 * nn - 1 - 2 * gdout.beg_c + gdin.beg_c
            pad_cd[:, 1] = 2 * nn - 2 + 2 * gdout.end_c - gdin.end_c
            neighborpad_cd[:, 0] = 2 * nn - 2 + 2 * gdout.beg_c - gdin.beg_c
            neighborpad_cd[:, 1] = 2 * nn - 1 - 2 * gdout.end_c + gdin.end_c
            self.interpolate = False
        else:
            assert (gdout.N_c == 2 * gdin.N_c).all()
            # Interpolation:
            pad_cd[:, 0] = nn - 1 - gdout.beg_c // 2 + gdin.beg_c
            pad_cd[:, 1] = nn + gdout.end_c // 2 - gdin.end_c
            neighborpad_cd[:, 0] = nn + gdout.beg_c // 2 - gdin.beg_c
            neighborpad_cd[:, 1] = nn - 1 - gdout.end_c // 2 + gdin.end_c
            skip_cd[:, 0] = gdout.beg_c % 2
            skip_cd[:, 1] = gdout.end_c % 2
            self.interpolate = True

            inpoints = (gdin.n_c[0] + 2 * nn - 1) * (gdin.n_c[1] + 2 * nn - 1)
            outpoints = gdout.n_c[0] * gdout.n_c[1]

            if inpoints > outpoints:
                points = ' x '.join([str(N) for N in gdin.N_c])
                raise ValueError('Cannot construct interpolator.  Grid %s '
                                 'may be too small' % points)

        assert (pad_cd.ravel() >= 0).all()
        self.ngpin = tuple(gdin.n_c)
        self.ngpout = tuple(gdout.n_c)
        assert dtype in [float, complex]

        self.pad_cd = pad_cd
        self.neighborpad_cd = neighborpad_cd
        self.skip_cd = skip_cd

        if gdin.comm.size > 1:
            comm = gdin.comm.get_c_object()
        else:
            comm = None

        self.transformer = _gpaw.Transformer(gdin.n_c, gdout.n_c,
                                             2 * nn, pad_cd,
                                             neighborpad_cd, skip_cd,
                                             gdin.neighbor_cd,
                                             dtype == float, comm,
                                             self.interpolate,
                                             xp is not np)

    def apply(self, input, output=None, phases=None):
        if output is None:
            output = self.gdout.empty(input.shape[:-3], dtype=self.dtype,
                                      xp=self.xp)
        if self.xp is np:
            self.transformer.apply(input, output, phases)
        else:
            self.transformer.apply_gpu(input.data.ptr,
                                       output.data.ptr,
                                       input.shape, input.dtype, phases)
        return output

    def get_async_sizes(self):
        return self.transformer.get_async_sizes()


class TransformerWrapper:
    def __init__(self, transformer):
        self.transformer = transformer
        self.dtype = transformer.dtype
        self.ngpin = transformer.ngpin
        self.ngpout = transformer.ngpout
        self.nn = transformer.nn

    def apply(self, input, output=None, phases=None):
        assert is_contiguous(input, self.dtype)
        assert input.shape[-3:] == self.ngpin
        if output is not None:
            assert is_contiguous(output, self.dtype)
            assert output.shape[-3:] == self.ngpout
        assert (self.dtype == float or
                (phases.dtype == complex and
                 phases.shape == (3, 2)))

        return self.transformer.apply(input, output, phases)

    def get_async_sizes(self):
        return self.transformer.get_async_sizes()


def Transformer(gdin, gdout, nn=1, dtype=float, xp=np):
    if nn != 9:
        t = _Transformer(gdin, gdout, nn, dtype, xp)
        if debug:
            t = TransformerWrapper(t)
        return t

    class T:
        nn = 1

        def apply(self, input, output, phases=None):
            output[:] = input

    return T()


def multiple_transform_apply(transformerlist, inputs, outputs, phases=None):
    return _gpaw.multiple_transform_apply(transformerlist, inputs, outputs,
                                          phases)


def coefs(k, p):
    for i in range(0, k * p, p):
        print('%2d' % i, end=' ')
        for x in range((k // 2 - 1) * p, k // 2 * p + 1):
            n = 1
            d = 1
            for j in range(0, k * p, p):
                if j == i:
                    continue
                n *= x - j
                d *= i - j
            print('%14.16f' % (n / d), end=' ')
        print()



from ase import Atoms
from gpaw import GPAW
import numpy as np

# spin paired H2


def test_xc_pygga():
    d = 0.75
    h2 = Atoms('H2', [[0, 0, 0], [0, 0, d]])
    h2.center(vacuum=2.)

    e = np.array([])
    f = np.array([])

    for XC in ['pyPBE', 'pyRPBE']:
        e = np.array([])
        f = np.array([])
        for i in [2, 0]:
            xc = XC[i:]
            calc = GPAW(nbands=-1, xc=xc,
                        h=0.3,
                        mode='lcao', basis='szp(dzp)')
            h2.calc = calc
            e = np.append(e, h2.get_potential_energy())
            f = np.append(f, h2.get_forces())
            del calc

        assert np.abs(e[0] - e[1]) < 1.e-4
        assert np.sum(np.abs(f[0] - f[1])) < 1.e-10

    # spin polarized O2
    d = 1.2
    o2 = Atoms('O2', [[0, 0, 0], [0, 0, d]], magmoms=[1., 1.])
    o2.center(vacuum=2.)

    for XC in ['pyPBE', 'pyRPBE']:
        e = np.array([])
        f = np.array([])
        for i in [2, 0]:
            xc = XC[i:]
            calc = GPAW(nbands=-2, xc=xc,
                        h=0.3,
                        mode='lcao', basis='szp(dzp)')
            o2.calc = calc
            e = np.append(e, o2.get_potential_energy())
            f = np.append(f, o2.get_forces())
            del calc

        assert np.abs(e[0] - e[1]) < 5.e-3
        assert np.sum(np.abs(f[0] - f[1])) < 1.e-4



"""Small test for local parts using yukawa potential"""
import pytest
from math import factorial as fac
from numpy import exp, sqrt, pi, absolute
from gpaw.atom.radialgd import AERadialGridDescriptor

# Values from Rico, Lopez, Ramirez, Ema, Theor Chem Acc (2013) 132:1304
#   Table 1 and 2.
#
#   Normalized STO: N = n + L : lambda (gamma) = 0.34
#   n    n'   L   Zeta     Zeta'      I(n,n',L,Zeta,Zeta')
#   8    6    4   8.4      3.5        0.793087136533672
#   2    1    5   8.4      3.5        0.271337849250231
#   3    4    6   8.4      3.5        0.170346973782779
#   3    2    8   8.4      3.5        0.141826311951003
#   2    1   10   8.4      3.5        9.005003835363871(-2)

#
#   Zeta = lambda (gamma)
#   n = 3, L=5, n'=2, L'=5, zeta'=3.5
#   0.10                              9.503545229396430(-6)
#   0.25                              5.869262722600324(-4)
#   0.50                              1.194846721531269(-2)
#   0.75                              5.829196062489732(-2)
#   1.00                              0.153003818405446


@pytest.fixture
def rgd():
    N = 1200
    beta = 0.4 * 600 / N
    return AERadialGridDescriptor(beta / N, 1.0 / N, N)


basic_sto_list = [  # See table 1 of Rico, Lopez, Ramirez, Ema,
                    # Theor Chem Acc (2013) 132:1304
    {
        'n': [8, 6],            # n, n'
        'Zeta': [8.4, 3.5],     # Z, Z'
        'L': 4,                 # L
        'I': 0.793087136533672  # [X_LM^n|X_LM^n']
    },
    # {
    #     'n': [2, 1],
    #     'Zeta': [8.4, 3.5],
    #     'L': 5,
    #     'I': 0.271337849250231
    # },
    # {          # values are completely of
    #     'n': [3, 4],
    #     'Zeta': [8.4, 3.5],
    #     'L': 6,
    #     'I': 0.170346973782779
    # },
    # {
    #     'n': [3, 2],            # n, n'
    #     'Zeta': [8.4, 3.5],     # Z, Z'
    #     'L': 8,                 # L
    #     'I': 0.141826311951003  # [X_LM^n|X_LM^n']
    # },
    # {
    #     'n': [2, 1],
    #     'Zeta': [8.4, 3.5],
    #     'L': 10,
    #     'I': 9.005003835363871e-2
    # }
]

gamma_sto_list = [  # See table 2 of Rico, Lopez, Ramirez, Ema,
                    # Theor Chem Acc (2013) 132:1304
    {
        'zeta': 0.10,
        'I': 9.503545229396430e-6
    },
    # {
    #     'zeta': 0.25,
    #     'I': 5.869262722600324e-4
    # },
    # {
    #     'zeta': 0.50,
    #     'I': 1.194846721531269e-2
    # },
    # {
    #     'zeta': 0.75,
    #     'I': 5.829196062489732e-2
    # },
    # {
    #     'zeta': 1.00,
    #     'I': 0.153003818405446
    # },
]


def radial_sto(n, zeta, l, r):
    """Build radial part of slater type orbital"""
#       Stolen from gpaw all_electron.py (intialize_wave_functions)
#
#       STOs are defined as
#
#           N Y_l^m r^{n-1} exp(-zeta * r)
#
#           N = (2 zeta)^n sqrt(2 zeta / (2n)!)
    assert n > 0
    radial = r**(n + l - 1.0)
    radial *= exp(-zeta * r)
    N = (2 * zeta)**(n + l) * sqrt(2 * zeta / fac(2 * (n + l)))
    radial *= N  # perhaps also do Y_l^m normalization
    return radial


def test_different_sto(rgd):
    """Check integration of two different STO."""
    gamma = 0.34
    for test_sto in basic_sto_list:
        wave_functions = []
        l_sto = test_sto['L']
        for n, zeta in zip(test_sto['n'], test_sto['Zeta']):
            radial = radial_sto(n, zeta, l_sto, rgd.r_g)
            norm = sqrt((2 * l_sto + 1) / (4 * pi))  # m = 0
            radial *= norm
            wave_functions.append(radial)
        yukawa_wf = rgd.yukawa(wave_functions[1], l_sto, gamma)
        test_int = rgd.integrate(wave_functions[0] * yukawa_wf, -1) / \
            (2 * l_sto + 1)
#        print(u"{:7.5e}||{:7.5e}||{:7.5e}".format(test_sto['I'], I,
#                absolute(I - test_sto['I']) * 100.0 / test_sto['I']))
        assert absolute(test_int - test_sto['I']) / test_sto['I'] < 0.001


def test_poisson(rgd):
    """Check integration against rgd.poisson."""
    gamma = 1e-5
    for test_sto in basic_sto_list:
        wave_functions = []
        l_sto = test_sto['L']
        for n, zeta in zip(test_sto['n'], test_sto['Zeta']):
            radial = radial_sto(n, zeta, l_sto, rgd.r_g)
            norm = sqrt((2 * l_sto + 1) / (4 * pi))  # m = 0
            radial *= norm
            wave_functions.append(radial)
        yukawa_wf = rgd.yukawa(wave_functions[1], l_sto, gamma)
        poisson_wf = rgd.poisson(wave_functions[1], l_sto)
        I_p = rgd.integrate(wave_functions[0] * poisson_wf, -1)
        I_y = rgd.integrate(wave_functions[0] * yukawa_wf, -1)
        assert absolute(I_y - I_p) / I_p < 5e-4


def test_same_sto(rgd):
    """Check integration of nearly identical STO."""
    n = 3
    l_sto = 5
    np = 2
    zetap = 3.5
    for test_sto in gamma_sto_list:
        zeta = test_sto['zeta']
        radial = radial_sto(n, zeta, l_sto, rgd.r_g)
        norm = sqrt((2 * l_sto + 1) / (4 * pi))  # m = 0
        radial *= norm
        radial2 = radial_sto(np, zetap, l_sto, rgd.r_g)
        radial2 *= norm
        yukawa_wf = rgd.yukawa(radial2, l_sto, zeta)
        test_int = rgd.integrate(radial * yukawa_wf, -1) / (2 * l_sto + 1)
        assert absolute(test_int - test_sto['I']) / test_sto['I'] < 0.001



from gpaw import GPAW, restart
from ase.build import bcc111
from gpaw.poisson import PoissonSolver
import numpy as np


def test_poisson_metallic_poisson(in_tmp_dir):
    slab = bcc111('Na', (1, 1, 2), vacuum=8)
    electrodes = ['both', 'single']
    charge = 0.05

    for electrode in electrodes:
        slab.calc = GPAW(
            mode='fd',
            xc='LDA', h=0.22,
            txt='metallic.txt',
            charge=charge,
            convergence={'density': 1e-1,
                         'energy': 1e-1,
                         'eigenstates': 1e-1},
            kpts=(2, 2, 1),
            poissonsolver=PoissonSolver(metallic_electrodes=electrode))

        _ = slab.get_potential_energy()
        phi0 = slab.calc.get_electrostatic_potential()
        assert np.all(abs(phi0[:, :, 0]) < 1e-10)
        if electrode == 'both':
            print(phi0[:, :, 0])
            print(phi0[:, :, 1])
            # The last zero boundary condition is implicit, so extrapolate
            d = phi0[:, :, -1] - phi0[:, :, -2]
            assert np.all(abs(phi0[:, :, -1] + d) < 1e-5)

        slab.calc.write('%s.gpw' % electrode)

        atoms, calc = restart('%s.gpw' % electrode, txt='restart.txt')
        phi02 = calc.get_electrostatic_potential()
        assert np.all(abs(phi02 - phi0) < 1e-10)



import pytest
from gpaw.utilities.elpa import LibElpa
from ase.build import molecule
from gpaw.mpi import world
from gpaw import GPAW, Mixer

pytestmark = pytest.mark.skipif(not LibElpa.have_elpa(),
                                reason='not LibElpa.have_elpa()')


def test_lcao_lcao_elpa():
    size = (world.size // 2, 2) if world.size > 1 else (1, 1)

    energies = []
    for use_elpa in [1, 0]:
        atoms = molecule('CH3CH2OH', vacuum=2.5)
        calc = GPAW(mode='lcao', basis='dzp',
                    h=0.25,
                    parallel=dict(sl_default=(size[0], size[1], 3),
                                  use_elpa=use_elpa),
                    mixer=Mixer(0.5, 5, 50.0))
        atoms.calc = calc
        E = atoms.get_potential_energy()
        energies.append(E)
    err = abs(energies[1] - energies[0])
    assert err < 1e-10, err



from math import pi

import numpy as np
from scipy.special import erf, erfc


class Ewald:
    """Class for calculating Ewald summations.

    'cell' is the unit cell in the cartesian basis.

    'G' is a normalized Ewald parameter. Large G results in fast
    real-space convergence but slow convergence in reciprocal space.
    This describes the width (in reciprocal space, and inverse width
    in real space) of the Gaussian shaped probe charge.

    Ng and Nl are lists specifying the number or nearest neighbors in
    sums over the reciprocal lattice and the real space lattice
    respectively.
    """

    def __init__(self, cell, G=5, Ng=[9, 9, 9], Nl=[3, 3, 3]):
        self.cell = cell
        self.Vcell = abs(np.linalg.det(cell))
        self.recip_cell = np.linalg.inv(self.cell).T
        self.Ng = Ng
        self.Nl = Nl
        self.G = G

        # Renormalize G to the longest lattice vector
        self.G /= np.sqrt((cell**2).sum(1).max())

    def get_wigner_seitz_radius(self, N):
        """Wigner-Seitz radius for N electrons in the unit cell."""
        return (3 * self.Vcell / (4 * pi * N))**(1 / 3)

    def get_sum_recip_ij(self, r_v, eps=1e-10):
        r"""The reciprocal space sum.

        ::

              -----  -x  i g.r
           pi \     e   e
          ---2 |   -------   , with x = g^2/(4 G^2)
          V G /       x
              -----
             g not 0
        """
        N = self.Ng
        b1, b2, b3 = self.recip_cell
        E_g = 0.
        for i in range(-N[0], N[0] + 1):
            for j in range(-N[1], N[1] + 1):
                for k in range(-N[2], N[2] + 1):
                    g_v = 2 * pi * (i * b1 + j * b2 + k * b3)
                    g2 = np.dot(g_v, g_v)
                    if g2 > eps:  # exclude g=0
                        x = .25 * g2 / self.G**2
                        E_g += np.exp(1.j * np.dot(r_v, g_v) - x) / x
        return pi / (self.Vcell * self.G**2) * E_g.real

    def get_sum_real_ij(self, r_v, eps=1e-5):
        r"""The real space sum.

        ::

            -----
            \     erfc( G [l-r| )
             |   --------------
            /         |l-r|
            -----
           l not 0

        Note: Add the l=0 term with erfc(r).
        """
        N = self.Nl
        a1, a2, a3 = self.cell
        E_r = 0.
        for i in range(-N[0], N[0] + 1):
            for j in range(-N[1], N[1] + 1):
                for k in range(-N[2], N[2] + 1):
                    l_v = i * a1 + j * a2 + k * a3
                    if np.linalg.norm(l_v) > eps:  # exclude l=0
                        lr = np.linalg.norm(l_v - r_v)
                        E_r += erfc(self.G * lr) / lr
        return E_r

    def get_electrostatic_potential(self, r, r_B, q_B, excludefroml0=None):
        r"""...

        Calculates the electrostatic potential at point r_i from point
        charges at {r_B} in a lattice using the Ewald summation.

        Charge neutrality is obtained by adding the homogenous density
        q_hom/V::

                        ---- ----'                             -
                        \    \         q_j            q_hom   /      1
          phi(r_i)  =    |    |   ---------------  +  -----   |dr ---------
                        /    /    |r_i - r_j + l|       V     |   |r - r_i|
                        ---- ----                             /
                         j    l

        r_B : matrix with the lattice basis (in cartesian coordinates).

        q_B : probe charges (in units of e).

        excludefroml0 : integer specifying if a point charge is not to
        be included in the central (l=0) unit cell. Used for Madelung
        constants.
        """
        E0 = 0.
        if excludefroml0 is None:
            excludefroml0 = np.zeros(len(q_B), dtype=int)
        if excludefroml0 in range(len(q_B)):
            i = excludefroml0
            excludefroml0 = np.zeros(len(q_B), dtype=int)
            excludefroml0[i] = 1

        assert sum(excludefroml0) <= 1

        for i, q in enumerate(q_B):  # potential from point charges
            rprime = r - r_B[i]
            absr = np.linalg.norm(rprime)
            E0 += q * self.get_sum_real_ij(rprime)
            E0 += q * self.get_sum_recip_ij(rprime)
            if excludefroml0[i]:  # if sum over l not 0
                if absr < 1e-14:
                    # lim r -> 0 erf(r G) / r = 2 * G / sqrt(pi)
                    E0 -= q * 2. * self.G / np.sqrt(pi)
                else:
                    E0 -= q * erf(absr * self.G) / absr
            else:  # if sum over all l
                E0 += q * erfc(absr * self.G) / absr

        # correct for compensating homogeneous background
        q_hom = -sum(q_B)
        E0 += q_hom * pi / (self.G**2 * self.Vcell)

        return E0


def madelung(cell):
    return Ewald(cell).get_electrostatic_potential(np.zeros(3), np.zeros(3),
                                                   [-1], 0)



from ase.build import molecule
from gpaw import GPAW, PW
from gpaw.mpi import world


def test_pw_h(in_tmp_dir):
    a = molecule('H', pbc=1)
    a.center(vacuum=2)

    comm = world.new_communicator([world.rank])
    e0 = 0.0
    a.calc = GPAW(mode=PW(250),
                  communicator=comm,
                  txt=None)
    e0 = a.get_potential_energy()
    e0 = world.sum(e0) / world.size

    a.calc = GPAW(mode=PW(250),
                  eigensolver='rmm-diis',
                  basis='szp(dzp)',
                  txt='%d.txt' % world.size)
    e = a.get_potential_energy()
    f = a.get_forces()
    assert abs(e - e0) < 7e-5, abs(e - e0)
    assert abs(f).max() < 1e-10, abs(f).max()



# Copyright (C) 2009-2012 P. Kluepfel and the CoMPaS group
# Please see the accompanying LICENSE file for further information.
#
# ============================================================================
#                         IMPORTANT INFORMATION
# ============================================================================
# The SIC module is developed by the CoMPaS Group, currently
# affiliated with the Science Institute of the University of Iceland.
#
# The last updates of our contribution to the GPAW code took place
# 01.11.2011 and a minor update on 16.01.2012. Since then there is no active
# development in the original GPAW repositories, but of course we continued
# work in our development version.
#
# If you are interested in running self-interaction corrected DFT
# calculations we strongly recommend you to read the LICENSE file carefully
# (especially regarding the FITNESS FOR A PARTICULAR PURPOSE of THIS version
# of the code). In case of questions, a persistent interest in running SIC
# calculations or if you want to support the necessary further development
# of SIC and the theory of orbital-density dependent energy functionals
# in GPAW you can reach the developers at:
#
#     peter-Dot-kluepfel-At-gmail-Dot-com
#
# For further information see:
#
#     http://compas.hinet.info
#
# ============================================================================
"""Perdew-Zunger SIC to DFT functionals.

Self-consistent minimization of self-interaction corrected
functionals (Perdew-Zunger).
"""

from math import pi

import numpy as np
from ase.units import Bohr, Hartree
from scipy.linalg import eigh

from gpaw.utilities.blas import mmmx
from gpaw.xc import XC
from gpaw.xc.functional import XCFunctional
from gpaw.poisson import PoissonSolver
from gpaw.transformers import Transformer
from gpaw.utilities import pack, unpack
from gpaw.lfc import LFC
import _gpaw


def matrix_exponential(G_nn, dlt):
    """Computes the matrix exponential of an antihermitian operator

       U = exp(i*dlt*G)

    G_nn: ndarray
        anti-hermitian (skew-symmetric) matrix.

    dlt: float
        scaling factor for G_nn.
    """

    ndim = G_nn.shape[1]
    w_n = np.zeros((ndim), dtype=float)

    V_nn = np.zeros((ndim, ndim), dtype=complex)
    O_nn = np.zeros((ndim, ndim), dtype=complex)
    if G_nn.dtype == complex:
        V_nn = 1j * G_nn.real - G_nn.imag
    else:
        V_nn = 1j * G_nn.real

    w_n, V_nn = eigh(V_nn)

    O_nn = np.diag(np.exp(1j * dlt * w_n))

    if G_nn.dtype == complex:
        U_nn = np.dot(V_nn, np.dot(O_nn, V_nn.T.conj())).copy()
    else:
        U_nn = np.dot(V_nn, np.dot(O_nn, V_nn.T.conj())).real.copy()

    return U_nn


def ortho(W_nn, maxerr=1E-10):
    """ Orthogonalizes the column vectors of a matrix by Symmetric
        Loewdin orthogonalization

    W_nn: ndarray
        unorthonormal matrix.

    maxerr: float
        maximum error for using explicit diagonalization.

    """

    ndim = np.shape(W_nn)[1]

    # overlap matrix
    O_nn = np.dot(W_nn, W_nn.T.conj())

    # check error in orthonormality
    err = np.sum(np.abs(O_nn - np.eye(ndim)))
    if (err < maxerr):
        # perturbative Symmetric-Loewdin
        X_nn = 1.5 * np.eye(ndim) - 0.5 * O_nn
    else:
        # diagonalization
        n_n = np.zeros(ndim, dtype=float)
        n_n, U_nn = eigh(O_nn)
        nsqrt_n = np.diag(1.0 / np.sqrt(n_n))
        X_nn = np.dot(np.dot(U_nn, nsqrt_n), U_nn.T.conj())

    # apply orthonormalizing transformation
    O_nn = np.dot(X_nn, W_nn)

    return O_nn


def random_unitary_matrix(delta, n):
    """ Initializaes a (random) unitary matrix

    delta: float
        strength of deviation from unit-matrix.
        > 0 random matrix
        < 0 non-random matrix

    n: int
        dimensionality of matrix.
    """

    assert n > 0

    # special case n=1:
    if n == 1:
        return np.identity(1)

    # non-random unitary matrix
    if delta < 0:
        W_nn = np.zeros((n, n))
        for i in range(n):
            for j in range(i):
                W_nn[i, j] = 1.0 / (i + j)
        W_nn = (W_nn - W_nn.T)

        return matrix_exponential(W_nn, delta)

    # random unitary matrix
    elif delta > 0:
        W_nn = np.random.rand(n, n)
        W_nn = (W_nn - W_nn.T)

        return matrix_exponential(W_nn, delta)
    else:
        return np.identity(n)


class SIC(XCFunctional):

    orbital_dependent = True
    unitary_invariant = False

    def __init__(self, xc='LDA', finegrid=False, **parameters):
        """Self-Interaction Corrected Functionals (PZ-SIC).

        finegrid: boolean
            Use fine grid for energy functional evaluations?
        """

        if isinstance(xc, str):
            xc = XC(xc)

        if xc.orbital_dependent:
            raise ValueError('SIC does not support ' + xc.name)

        self.xc = xc
        XCFunctional.__init__(self, xc.name + '-PZ-SIC', xc.type)
        self.finegrid = finegrid
        self.parameters = parameters

    # Proper IO of general SIC objects should work by means of something like:
    def todict(self):
        return {
            'type': 'SIC',
            'name': self.xc.name,
            'finegrid': self.finegrid,
            'parameters': dict(self.parameters)}

    def initialize(self, density, hamiltonian, wfs):
        assert wfs.kd.gamma
        assert not wfs.gd.pbc_c.any()
        assert wfs.bd.comm.size == 1  # band parallelization unsupported

        self.wfs = wfs
        self.dtype = float
        self.xc.initialize(density, hamiltonian, wfs)
        self.kpt_comm = wfs.kd.comm
        self.nspins = wfs.nspins
        self.nbands = wfs.bd.nbands

        self.finegd = density.gd.refine() if self.finegrid else density.gd
        self.ghat = LFC(self.finegd,
                        [setup.ghat_l for setup in density.setups],
                        integral=np.sqrt(4 * np.pi),
                        forces=True)

        # XXX Use hamiltonian.poisson instead?
        poissonsolver = PoissonSolver()
        poissonsolver.set_grid_descriptor(self.finegd)

        self.spin_s = {}
        for kpt in wfs.kpt_u:
            self.spin_s[kpt.s] = SICSpin(kpt, self.xc, density, hamiltonian,
                                         wfs, poissonsolver, self.ghat,
                                         self.finegd, **self.parameters)

    def get_setup_name(self):
        return self.xc.get_setup_name()

    def calculate_paw_correction(self,
                                 setup,
                                 D_sp,
                                 dEdD_sp=None,
                                 addcoredensity=True,
                                 a=None):
        return self.xc.calculate_paw_correction(setup, D_sp, dEdD_sp,
                                                addcoredensity, a)

    def set_positions(self, spos_ac):
        self.ghat.set_positions(spos_ac)

    def calculate(self, gd, n_sg, v_sg=None, e_g=None):
        # Normal XC contribution:
        exc = self.xc.calculate(gd, n_sg, v_sg, e_g)

        # SIC:
        self.esic = 0.0
        self.ekin = 0.0
        for spin in self.spin_s.values():
            if spin.kpt.psit_nG is not None:
                desic, dekin = spin.calculate()
                self.esic += desic
                self.ekin += dekin
        self.esic = self.kpt_comm.sum(self.esic)
        self.ekin = self.kpt_comm.sum(self.ekin)

        return exc + self.esic

    def apply_orbital_dependent_hamiltonian(self,
                                            kpt,
                                            psit_nG,
                                            Htpsit_nG=None,
                                            dH_asp=None):
        spin = self.spin_s[kpt.s]
        if spin.W_mn is None:
            return
        spin.apply_orbital_dependent_hamiltonian(psit_nG)

    def correct_hamiltonian_matrix(self, kpt, H_nn):
        spin = self.spin_s[kpt.s]
        if spin.W_mn is None:
            return
        spin.correct_hamiltonian_matrix(H_nn)

    def add_correction(self,
                       kpt,
                       psit_xG,
                       Htpsit_xG,
                       P_axi,
                       c_axi,
                       n_x,
                       calculate_change=False):
        spin = self.spin_s[kpt.s]
        if spin.W_mn is None:
            return

        if calculate_change:
            spin.calculate_residual_change(psit_xG, Htpsit_xG, P_axi, c_axi,
                                           n_x)
        else:
            spin.calculate_residual(psit_xG, Htpsit_xG, P_axi, c_axi)

    def rotate(self, kpt, U_nn):
        self.spin_s[kpt.s].rotate(U_nn)

    def setup_force_corrections(self, F_av):
        self.dF_av = np.zeros_like(F_av)
        for spin in self.spin_s.values():
            spin.add_forces(self.dF_av)
        self.wfs.kd.comm.sum(self.dF_av)

    def add_forces(self, F_av):
        F_av += self.dF_av

    def summary(self, log):
        for s in range(self.nspins):
            if s in self.spin_s:
                stabpot = self.spin_s[s].stabpot
                spin = self.spin_s[s]
                pos_mv = spin.get_centers()
                exc_m = spin.exc_m
                ecoulomb_m = spin.ecoulomb_m
                if self.kpt_comm.rank == 1 and self.finegd.comm.rank == 0:
                    nocc = self.kpt_comm.sum(spin.nocc)
                    self.kpt_comm.send(pos_mv, 0)
                    self.kpt_comm.send(exc_m, 0)
                    self.kpt_comm.send(ecoulomb_m, 0)
            else:
                if self.kpt_comm.rank == 0 and self.finegd.comm.rank == 0:
                    nocc = self.kpt_comm.sum(0)
                    pos_mv = np.zeros((nocc, 3))
                    exc_m = np.zeros(nocc)
                    ecoulomb_m = np.zeros(nocc)
                    self.kpt_comm.receive(pos_mv, 1)
                    self.kpt_comm.receive(exc_m, 1)
                    self.kpt_comm.receive(ecoulomb_m, 1)
            if self.kpt_comm.rank == 0 and self.finegd.comm.rank == 0:
                log('\nSIC orbital centers and energies:')
                log('                                %5.2fx   %5.2fx' %
                    (self.spin_s[0].xc_factor, self.spin_s[0].coulomb_factor))
                log('          x       y       z       XC    Coulomb')
                log('--------------------------------------------------')
                m = 0
                for pos_v, exc, ecoulomb in zip(pos_mv, exc_m, ecoulomb_m):
                    log('%3d  (%7.3f,%7.3f,%7.3f): %8.3f %8.3f' %
                        ((m, ) + tuple(pos_v) +
                         (exc * Hartree, ecoulomb * Hartree)))
                    m += 1
                log('--------------------------------------------------')
        log('\nTotal SIC energy     : %12.5f' % (self.esic * Hartree))
        log('Stabilizing potential: %12.5f' % (stabpot * Hartree))

    def read(self, reader):
        xc_factor = reader.hamiltonian.xc.sic_xc_factor
        coulomb_factor = reader.hamiltonian.xc.sic_coulomb_factor

        for s in range(self.nspins):
            W_mn = reader.hamiltonian.xc.get(
                'unitary_transformation{0}'.format(s))

            if s in self.spin_s:
                self.spin_s[s].initial_W_mn = W_mn
                self.spin_s[s].xc_factor = xc_factor
                self.spin_s[s].coulomb_factor = coulomb_factor

    def write(self, writer):
        for s in self.spin_s:
            spin = self.spin_s[s]
            writer.write(
                sic_xc_factor=spin.xc_factor,
                sic_coulomb_factor=spin.coulomb_factor)
            break

        for s in range(self.nspins):
            W_mn = self.get_unitary_transformation(s)

            if W_mn is not None:
                writer.write('unitary_transformation{0}'.format(s), W_mn)

    def get_unitary_transformation(self, s):
        if s in self.spin_s.keys():
            spin = self.spin_s[s]

            if spin.W_mn is None or spin.finegd.rank != 0:
                n = 0
            else:
                n = spin.W_mn.shape[0]
        else:
            n = 0

        n = self.wfs.world.sum(n)

        if n > 0:
            W_mn = np.zeros((n, n), dtype=self.dtype)
        else:
            W_mn = None
            return W_mn

        if s in self.spin_s.keys():
            spin = self.spin_s[s]
            #
            if spin.W_mn is None or spin.finegd.rank != 0:
                W_mn[:] = 0.0
            else:
                W_mn[:] = spin.W_mn[:]
            #
        else:
            W_mn[:] = 0.0

        self.wfs.world.sum(W_mn)
        return W_mn


class SICSpin:
    def __init__(self,
                 kpt,
                 xc,
                 density,
                 hamiltonian,
                 wfs,
                 poissonsolver,
                 ghat,
                 finegd,
                 dtype=float,
                 coulomb_factor=0.5,
                 xc_factor=0.5,
                 uominres=1E-1,
                 uomaxres=1E-10,
                 uorelres=1E-4,
                 uonscres=1E-10,
                 rattle=-0.1,
                 stabpot=0.0,
                 maxuoiter=10,
                 logging=2):
        """Single spin SIC object.


        coulomb_factor:
            Scaling factor for Hartree-functional

        xc_factor:
            Scaling factor for xc-functional

        uominres:
            Minimum residual before unitary optimization starts

        uomaxres:
            Target accuracy for unitary optimization
            (absolute variance)

        uorelres:
            Target accuracy for unitary optimization
            (rel. to basis residual)

        maxuoiter:
            Maximum number of unitary optimization steps

        rattle:
            perturbation to the initial states
        """

        self.wfs = wfs
        self.kpt = kpt
        self.xc = xc
        self.poissonsolver = poissonsolver
        self.ghat = ghat
        self.pt = wfs.pt

        self.gd = wfs.gd
        self.finegd = finegd

        if self.finegd is self.gd:
            self.interpolator = None
            self.restrictor = None
        else:
            self.interpolator = Transformer(self.gd, self.finegd, 3)
            self.restrictor = Transformer(self.finegd, self.gd, 3)

        self.nspins = wfs.nspins
        self.spin = kpt.s
        self.timer = wfs.timer
        self.setups = wfs.setups

        self.dtype = dtype
        self.coulomb_factor = coulomb_factor
        self.xc_factor = xc_factor

        self.nocc = None  # number of occupied states
        self.W_mn = None  # unit. transf. to energy optimal states
        self.initial_W_mn = None  # initial unitary transformation
        self.vt_mG = None  # orbital dependent potentials
        self.exc_m = None  # PZ-SIC contributions (from E_xc)
        self.ecoulomb_m = None  # PZ-SIC contributions (from E_H)

        self.rattle = rattle  # perturb the initial unitary transformation
        self.stabpot = stabpot  # stabilizing constant potential to avoid
        # occupation of unoccupied states

        self.uominres = uominres
        self.uomaxres = uomaxres
        self.uorelres = uorelres
        self.uonscres = uonscres
        self.maxuoiter = maxuoiter
        self.maxlsiter = 1  # maximum number of line-search steps
        self.maxcgiter = 2  # maximum number of CG-iterations
        self.lsinterp = True  # interpolate for minimum during line search
        self.basiserror = 1E+20
        self.logging = logging

    def initialize_orbitals(self, rattle=-0.1, localize=True):
        if self.initial_W_mn is not None:
            self.nocc = self.initial_W_mn.shape[0]
        else:
            self.nocc, x = divmod(int(self.kpt.f_n.sum()), 3 - self.nspins)
            assert x == 0

        if self.nocc == 0:
            return

        Z_mmv = np.empty((self.nocc, self.nocc, 3), complex)
        for v in range(3):
            G_v = np.zeros(3)
            G_v[v] = 1
            Z_mmv[:, :, v] = self.gd.wannier_matrix(
                self.kpt.psit_nG, self.kpt.psit_nG, G_v, self.nocc)
        self.finegd.comm.sum(Z_mmv)

        if self.initial_W_mn is not None:
            self.W_mn = self.initial_W_mn

        elif localize:
            W_nm = np.identity(self.nocc)
            localization = 0.0
            for iter in range(30):
                loc = _gpaw.localize(Z_mmv, W_nm)
                if loc - localization < 1e-6:
                    break
                localization = loc

            self.W_mn = W_nm.T.copy()
        else:
            self.W_mn = np.identity(self.nocc)

        if (rattle != 0.0 and self.W_mn is not None and
                self.initial_W_mn is None):
            U_mm = random_unitary_matrix(rattle, self.nocc)
            self.W_mn = np.dot(U_mm, self.W_mn)

        if self.W_mn is not None:
            self.finegd.comm.broadcast(self.W_mn, 0)

        spos_mc = -np.angle(Z_mmv.diagonal()).T / (2 * pi)
        self.initial_pos_mv = np.dot(spos_mc % 1.0, self.gd.cell_cv)

    def localize_orbitals(self):

        assert self.gd.orthogonal

        # calculate wannier matrixelements
        Z_mmv = np.empty((self.nocc, self.nocc, 3), complex)
        for v in range(3):
            G_v = np.zeros(3)
            G_v[v] = 1
            Z_mmv[:, :, v] = self.gd.wannier_matrix(
                self.kpt.psit_nG, self.kpt.psit_nG, G_v, self.nocc)
        self.finegd.comm.sum(Z_mmv)

        # setup the initial configuration (identity)
        W_nm = np.identity(self.nocc)

        # localize the orbitals
        localization = 0.0
        for iter in range(30):
            loc = _gpaw.localize(Z_mmv, W_nm)
            if loc - localization < 1e-6:
                break
            localization = loc

        # apply localizing transformation
        self.W_mn = W_nm.T.copy()

    def rattle_orbitals(self, rattle=-0.1):

        # check for the trivial cases
        if rattle == 0.0:
            return

        if self.W_mn is None:
            return

        # setup a "random" unitary matrix
        nocc = self.W_mn.shape[0]
        U_mm = random_unitary_matrix(rattle, nocc)

        # apply unitary transformation
        self.W_mn = np.dot(U_mm, self.W_mn)

    def get_centers(self):
        assert self.gd.orthogonal

        # calculate energy optimal states (if necessary)
        if self.phit_mG is None:
            self.update_optimal_states()

        # calculate wannier matrixelements
        Z_mmv = np.empty((self.nocc, self.nocc, 3), complex)
        for v in range(3):
            G_v = np.zeros(3)
            G_v[v] = 1
            Z_mmv[:, :, v] = self.gd.wannier_matrix(self.phit_mG, self.phit_mG,
                                                    G_v, self.nocc)
        self.finegd.comm.sum(Z_mmv)

        # calculate positions of localized orbitals
        spos_mc = -np.angle(Z_mmv.diagonal()).T / (2 * pi)

        return np.dot(spos_mc % 1.0, self.gd.cell_cv) * Bohr

    def calculate_sic_matrixelements(self):
        # overlap of pseudo wavefunctions
        Htphit_mG = self.vt_mG * self.phit_mG
        V_mm = np.zeros((self.nocc, self.nocc), dtype=self.dtype)
        # gemm(self.gd.dv, self.phit_mG, Htphit_mG, 0.0, V_mm, 't')
        mmmx(self.gd.dv,
             Htphit_mG, 'N',
             self.phit_mG, 'T', 0.0, V_mm)

        # PAW
        for a, P_mi in self.P_ami.items():
            for m, dH_p in enumerate(self.dH_amp[a]):
                dH_ii = unpack(dH_p)
                V_mm[m, :] += np.dot(P_mi[m], np.dot(dH_ii, P_mi.T))

        # accumulate over grid-domains
        self.finegd.comm.sum(V_mm)
        self.V_mm = V_mm

        # Symmetrization of V and kappa-matrix:
        K_mm = 0.5 * (V_mm - V_mm.T.conj())
        V_mm = 0.5 * (V_mm + V_mm.T.conj())

        # evaluate the kinetic correction
        self.ekin = -np.trace(V_mm) * (3 - self.nspins)

        return V_mm, K_mm, np.vdot(K_mm, K_mm).real

    def update_optimal_states(self):
        # pseudo wavefunctions
        self.phit_mG = self.gd.zeros(self.nocc)
        if self.nocc > 0:
            mmmx(1.0, self.W_mn, 'N', self.kpt.psit_nG[:self.nocc], 'N', 0.0,
                 self.phit_mG)

        # PAW
        self.P_ami = {}
        for a, P_ni in self.kpt.P_ani.items():
            self.P_ami[a] = np.dot(self.W_mn, P_ni[:self.nocc])

    def calculate_density(self, m):
        # pseudo density
        nt_G = self.phit_mG[m]**2

        if self.finegd is self.gd:
            nt_g = nt_G
        else:
            nt_g = self.finegd.empty()
            self.interpolator.apply(nt_G, nt_g)
            # normalize single-particle density
            nt_g *= self.gd.integrate(nt_G) / self.finegd.integrate(nt_g)

        # PAW corrections
        Q_aL = {}
        D_ap = {}
        for a, P_mi in self.P_ami.items():
            P_i = P_mi[m]
            D_ii = np.outer(P_i, P_i.conj()).real
            D_ap[a] = D_p = pack(D_ii)
            Q_aL[a] = np.dot(D_p, self.setups[a].Delta_pL)

        return nt_g, Q_aL, D_ap

    def update_potentials(self, save=False, restore=False):
        if restore:
            self.exc_m = self.exc_save_m
            self.ecoulomb_m = self.eco_save_m
            self.esic = self.esic_save
            self.vt_mG = self.vt_save_mG.copy()
            self.dH_amp = self.dH_save_amp.copy()
            return

        self.timer.start('ODD-potentials')

        nt_sg = self.finegd.empty(2)
        nt_sg[1] = 0.0
        vt_sg = self.finegd.empty(2)

        # PAW
        W_aL = self.ghat.dict()
        zero_initial_phi = False

        # initialize some bigger fields
        if self.vt_mG is None or self.nocc != self.phit_mG.shape[0]:
            self.vt_mG = self.gd.empty(self.nocc)
            self.exc_m = np.zeros(self.nocc)
            self.ecoulomb_m = np.zeros(self.nocc)
            self.vHt_mg = self.finegd.zeros(self.nocc)
            zero_initial_phi = True
        #
        # PAW
        self.dH_amp = {}
        for a, P_mi in self.P_ami.items():
            ni = P_mi.shape[1]
            self.dH_amp[a] = np.empty((self.nocc, ni * (ni + 1) // 2))
        #
        self.Q_maL = {}
        # loop all occupied orbitals
        for m, phit_G in enumerate(self.phit_mG):
            #
            # setup the single-particle density and PAW density-matrix
            nt_sg[0], Q_aL, D_ap = self.calculate_density(m)
            vt_sg[:] = 0.0
            #
            # xc-SIC
            self.timer.start('XC')
            if self.xc_factor != 0.0:
                exc = self.xc.calculate(self.finegd, nt_sg, vt_sg)
                exc /= self.finegd.comm.size
                vt_sg[0] *= -self.xc_factor
                #
                # PAW
                for a, D_p in D_ap.items():
                    setup = self.setups[a]
                    dH_p = self.dH_amp[a][m]
                    dH_sp = np.zeros((2, len(dH_p)))
                    #
                    D_sp = np.array([D_p, np.zeros_like(D_p)])
                    exc += self.xc.calculate_paw_correction(
                        setup, D_sp, dH_sp, addcoredensity=False)
                    dH_p[:] = -dH_sp[0] * self.xc_factor

                self.exc_m[m] = -self.xc_factor * exc
            self.timer.stop('XC')
            #
            # Hartree-SIC
            self.timer.start('Hartree')
            if self.coulomb_factor != 0.0:
                #
                # add compensation charges to pseudo density
                self.ghat.add(nt_sg[0], Q_aL)
                #
                # solve the coulomb problem
                self.poissonsolver.solve(
                    self.vHt_mg[m],
                    nt_sg[0],
                    zero_initial_phi=zero_initial_phi)
                ecoulomb = 0.5 * self.finegd.integrate(nt_sg[0] *
                                                       self.vHt_mg[m])
                ecoulomb /= self.finegd.comm.size
                vt_sg[0] -= self.coulomb_factor * self.vHt_mg[m]
                #
                # PAW
                self.ghat.integrate(self.vHt_mg[m], W_aL)
                for a, D_p in D_ap.items():
                    setup = self.setups[a]
                    dH_p = self.dH_amp[a][m]
                    M_p = np.dot(setup.M_pp, D_p)
                    ecoulomb += np.dot(D_p, M_p)
                    #
                    dH_p -= self.coulomb_factor * (
                        2.0 * M_p + np.dot(setup.Delta_pL, W_aL[a]))
                #
                self.ecoulomb_m[m] = -self.coulomb_factor * ecoulomb
            self.timer.stop('Hartree')

            # restrict to course grid
            if self.finegd is self.gd:
                self.vt_mG[m] = vt_sg[0]
            else:
                self.timer.start('Restrictor')
                self.restrictor.apply(vt_sg[0], self.vt_mG[m])
                self.timer.stop('Restrictor')
            self.Q_maL[m] = Q_aL

        self.timer.stop('ODD-potentials')

        # accumulate total xc-SIC and coulomb-SIC
        self.finegd.comm.sum(self.exc_m)
        self.finegd.comm.sum(self.ecoulomb_m)

        # total sic (including spin-degeneracy)
        self.esic = self.exc_m.sum()
        self.esic += self.ecoulomb_m.sum()
        self.esic *= (3 - self.nspins)

        if save:
            self.exc_save_m = self.exc_m
            self.eco_save_m = self.ecoulomb_m
            self.esic_save = self.esic
            self.vt_save_mG = self.vt_mG.copy()
            self.dH_save_amp = self.dH_amp.copy()

    def apply_orbital_dependent_hamiltonian(self, psit_nG):
        """...

        Setup::

            |V phi_m> and <l|Vphi_m>,

        for occupied states m and unoccupied states l."""

        # nocc = self.nocc
        # nvirt = psit_nG.shape[0] - nocc

        self.Htphit_mG = self.vt_mG * self.phit_mG

    def correct_hamiltonian_matrix(self, H_nn):
        """ Add contributions of the non-local part of the
            interaction potential to the Hamiltonian matrix.

            on entry:
                          H_nn[n,m] = <n|H_{KS}|m>
            on exit:
                          H_nn[n,m] = <n|H_{KS} + V_{u}|m>

            where V_u is the unified Hamiltonian

                V_u = ...

        """

        nocc = self.nocc
        nvirt = H_nn.shape[0] - nocc

        W_mn = self.W_mn
        # R_mk = self.R_mk

        if self.gd.comm.rank == 0:
            V_mm = 0.5 * (self.V_mm + self.V_mm.T)
            H_nn[:nocc, :nocc] += np.dot(W_mn.T, np.dot(V_mm, W_mn))
            if self.stabpot != 0.0:
                H_nn[nocc:, nocc:] += np.eye(nvirt) * self.stabpot

        if nvirt != 0:
            H_nn[:nocc, nocc:] = 0.0  # R_nk
            H_nn[nocc:, :nocc] = 0.0  # R_nk.T
            # R_nk = np.dot(W_mn.T, R_mk) # CHECK THIS
            # H_nn[:nocc, nocc:] += R_nk
            # H_nn[nocc:, :nocc] += R_nk.T

    def calculate_residual(self, psit_nG, Htpsit_nG, P_ani, c_ani):
        """ Calculate the action of the unified Hamiltonian on an
            arbitrary state:

                H_u|Psi> =
        """

        nocc = self.nocc
        nvirt = psit_nG.shape[0] - nocc

        # constraint for unoccupied states
        R_mk = np.zeros((nocc, nvirt), dtype=self.dtype)
        if nvirt > 0:
            mmmx(self.gd.dv, self.Htphit_mG, 'N', psit_nG[nocc:], 'T',
                 0.0, R_mk)
            # PAW
            for a, P_mi in self.P_ami.items():
                P_ni = P_ani[a]

                for m, dH_p in enumerate(self.dH_amp[a]):
                    dH_ii = unpack(dH_p)
                    R_mk[m] += np.dot(P_mi[m], np.dot(dH_ii, P_ni[nocc:].T))

            self.finegd.comm.sum(R_mk)

        # self.R_mk = R_mk
        # R_mk  = self.R_mk
        W_mn = self.W_mn
        Htphit_mG = self.Htphit_mG
        phit_mG = self.phit_mG
        K_mm = 0.5 * (self.V_mm - self.V_mm.T)
        Q_mn = np.dot(K_mm, W_mn)

        # Action of unified Hamiltonian on occupied states:
        if nocc > 0:
            mmmx(1.0, W_mn.T.copy(), 'N', Htphit_mG, 'N',
                 1.0, Htpsit_nG[:nocc])
            mmmx(1.0, Q_mn.T.copy(), 'N', phit_mG, 'N', 1.0, Htpsit_nG[:nocc])
        if nvirt > 0:
            mmmx(1.0, R_mk.T.copy(), 'N', phit_mG, 'N', 1.0, Htpsit_nG[nocc:])
            if self.stabpot != 0.0:
                Htpsit_nG[nocc:] += self.stabpot * psit_nG[nocc:]

        # PAW
        for a, P_mi in self.P_ami.items():
            #
            c_ni = c_ani[a]
            ct_mi = P_mi.copy()
            #
            dO_ii = self.setups[a].dO_ii
            c_ni[:nocc] += np.dot(Q_mn.T, np.dot(P_mi, dO_ii))
            c_ni[nocc:] += np.dot(R_mk.T, np.dot(P_mi, dO_ii))
            #
            for m, dH_p in enumerate(self.dH_amp[a]):
                dH_ii = unpack(dH_p)
                ct_mi[m] = np.dot(P_mi[m], dH_ii)
            c_ni[:nocc] += np.dot(W_mn.T, ct_mi)
            c_ni[nocc:] += self.stabpot * np.dot(P_ani[a][nocc:], dO_ii)

    def calculate_residual_change(self, psit_xG, Htpsit_xG, P_axi, c_axi, n_x):
        """

        """
        assert len(n_x) == 1

        Htphit_mG = self.Htphit_mG
        phit_mG = self.phit_mG

        w_mx = np.zeros((self.nocc, 1), dtype=self.dtype)
        v_mx = np.zeros((self.nocc, 1), dtype=self.dtype)

        mmmx(self.gd.dv, phit_mG, 'N', psit_xG, 'T', 0.0, w_mx)
        mmmx(self.gd.dv, Htphit_mG, 'N', psit_xG, 'T', 0.0, v_mx)

        # PAW
        for a, P_mi in self.P_ami.items():
            P_xi = P_axi[a]
            dO_ii = self.setups[a].dO_ii
            #
            w_mx += np.dot(P_mi, np.dot(dO_ii, P_xi.T))

            for m, dH_p in enumerate(self.dH_amp[a]):
                dH_ii = unpack(dH_p)
                v_mx[m] += np.dot(P_mi[m], np.dot(dH_ii, P_xi.T))

        # sum over grid-domains
        self.finegd.comm.sum(w_mx)
        self.finegd.comm.sum(v_mx)

        V_mm = 0.5 * (self.V_mm + self.V_mm.T)
        q_mx = v_mx - np.dot(V_mm, w_mx)

        if self.stabpot != 0.0:
            q_mx -= self.stabpot * w_mx

        mmmx(1.0, w_mx.T.copy(), 'N', Htphit_mG, 'N', 1.0, Htpsit_xG)
        mmmx(1.0, q_mx.T.copy(), 'N', phit_mG, 'N', 1.0, Htpsit_xG)

        # PAW
        for a, P_mi in self.P_ami.items():
            c_xi = c_axi[a]
            ct_mi = P_mi.copy()

            dO_ii = self.setups[a].dO_ii
            c_xi += np.dot(q_mx.T, np.dot(P_mi, dO_ii))

            for m, dH_p in enumerate(self.dH_amp[a]):
                dH_ii = unpack(dH_p)
                ct_mi[m] = np.dot(P_mi[m], dH_ii)
            c_xi += np.dot(w_mx.T, ct_mi)

        if self.stabpot != 0.0:
            Htphit_mG += self.stabpot * psit_xG
            for a, P_xi in P_axi.items():
                dO_ii = self.setups[a].dO_ii
                c_axi[a] += self.stabpot * np.dot(P_xi, dO_ii)

    def rotate(self, U_nn):
        """ Unitary transformations amongst the canonic states
            require to apply a counter-acting transformation to
            the energy optimal states. This subroutine takes
            care of it.

            Reorthogonalization is required whenever unoccupied
            states are mixed in.
        """
        # skip if no transformation to optimal states is set-up
        if self.W_mn is None:
            return

        # compensate the transformation amongst the occupied states
        self.W_mn = np.dot(self.W_mn, U_nn[:self.nocc, :self.nocc])

        # reorthogonalize if unoccupied states may have been mixed in
        if self.nocc != U_nn.shape[0]:
            self.W_mn = ortho(self.W_mn)
            # self.R_mk = np.dot(self.R_mk, U_nn[self.nocc:, self.nocc:].T)

    def add_forces(self, F_av):
        # Calculate changes in projections
        deg = 3 - self.nspins
        F_amiv = self.pt.dict(self.nocc, derivative=True)
        self.pt.derivative(self.phit_mG, F_amiv)
        for m in range(self.nocc):
            # Force from compensation charges:
            dF_aLv = self.ghat.dict(derivative=True)
            self.ghat.derivative(self.vHt_mg[m], dF_aLv)
            for a, dF_Lv in dF_aLv.items():
                F_av[a] -= deg * self.coulomb_factor * \
                    np.dot(self.Q_maL[m][a], dF_Lv)

            # Force from projectors
            for a, F_miv in F_amiv.items():
                F_vi = F_miv[m].T.conj()
                dH_ii = unpack(self.dH_amp[a][m])
                P_i = self.P_ami[a][m]
                F_v = np.dot(np.dot(F_vi, dH_ii), P_i)
                F_av[a] += deg * 2 * F_v.real

    def calculate(self):
        """ Evaluate the non-unitary invariant part of the
            energy functional and returns

            esic: float
                self-interaction energy

            ekin: float
                correction to the kinetic energy
        """
        # initialize transformation from canonic to energy
        # optimal states (if necessary)
        if self.W_mn is None:
            self.initialize_orbitals(rattle=self.rattle)

        # optimize the non-unitary invariant part of the
        # functional
        self.unitary_optimization()

        return self.esic, self.ekin

    def unitary_optimization(self):
        """ Optimization of the non-unitary invariant part of the
            energy functional.
        """

        optstep = 0.0
        Gold = 0.0
        cgiter = 0
        #
        epsstep = 0.005  # 0.005
        dltstep = 0.1  # 0.1
        prec = 1E-7

        # get the initial ODD potentials/energy/matrixelements
        self.update_optimal_states()
        self.update_potentials(save=True)
        ESI = self.esic
        V_mm, K_mm, norm = self.calculate_sic_matrixelements()

        if norm < self.uonscres and self.maxuoiter > 0:
            return

        if self.nocc <= 1:
            return

        # optimize the unitary transformation
        #
        # allocate arrays for the search direction,
        # i.e., the (conjugate) gradient
        D_old_mm = np.zeros_like(self.W_mn)

        for iter in range(abs(self.maxuoiter)):
            # copy the initial unitary transformation and orbital
            # dependent energies
            W_old_mn = self.W_mn.copy()

            # setup the steepest-descent/conjugate gradient
            # D_nn:  search direction
            # K_nn:  inverse gradient
            # G0  :  <K,D> (projected length of D along K)
            if Gold != 0.0:
                # conjugate gradient
                G0 = np.sum(K_mm * K_mm.conj()).real
                beta = G0 / Gold
                Gold = G0
                D_mm = K_mm + beta * D_old_mm

                G0 = np.sum(K_mm * D_mm.conj()).real
            else:
                # steepest-descent
                G0 = np.sum(K_mm * K_mm.conj()).real
                Gold = G0
                D_mm = K_mm

            updated = False
            minimum = False
            failed = True
            noise = False
            E0 = ESI

            # try to estimate optimal step-length from change in length
            # of the gradient (force-only)
            if (epsstep != 0.0 and norm > self.uomaxres):
                step = epsstep
                while (True):
                    U_mm = matrix_exponential(D_mm, step)
                    self.W_mn = np.dot(U_mm, W_old_mn)
                    self.update_optimal_states()
                    self.update_potentials()
                    E1 = self.esic
                    K0_mm = K_mm.copy()
                    V_mm, K_mm, norm = self.calculate_sic_matrixelements()

                    # projected length of the gradient at the new position
                    G1 = np.sum(((K_mm - K0_mm) / step) * D_mm.conj()).real
                    #
                    if (abs(E1 - E0) < prec and E1 >= E0):
                        #
                        eps_works = True
                        Eeps = E1
                        noise = True
                    elif (E1 < E0):
                        #
                        # trial step reduced energy
                        eps_works = True
                        Eeps = E1
                    else:
                        #
                        # epsilon step did not work
                        eps_works = False
                        optstep = 0.0
                        break

                    # compute the optimal step size
                    # optstep = step*G0/(G0-G1)
                    # print -G0/G1
                    optstep = -G0 / G1

                    if (eps_works):
                        break

                    # print eps_works, optstep, G0/((G0-G1)/step)

                    # decide on the method for stepping
                if (optstep > 0.0):

                    # convex region -> force only estimate for minimum
                    U_mm = matrix_exponential(D_mm, optstep)
                    self.W_mn = np.dot(U_mm, W_old_mn)
                    self.update_optimal_states()
                    self.update_potentials()
                    E1 = self.esic
                    if (abs(E1 - E0) < prec and E1 >= E0):
                        V_mm, K_mm, norm = self.calculate_sic_matrixelements()
                        ESI = E1
                        optstep = optstep
                        lsiter = 0
                        maxlsiter = -1
                        updated = True
                        minimum = True
                        failed = False
                        lsmethod = 'CV-N'
                        noise = True
                    elif (E1 < E0):
                        V_mm, K_mm, norm = self.calculate_sic_matrixelements()
                        ESI = E1
                        optstep = optstep
                        lsiter = 0
                        maxlsiter = -1
                        updated = True
                        minimum = True
                        failed = False
                        lsmethod = 'CV-S'
                    else:
                        # self.K_unn[q] = K_nn
                        ESI = E0
                        step = optstep
                        optstep = 0.0
                        lsiter = 0
                        maxlsiter = self.maxlsiter
                        updated = False
                        minimum = False
                        failed = True
                        lsmethod = 'CV-F-CC'
                else:
                    # self.K_unn[q] = K_nn
                    ESI = E0
                    step = optstep
                    optstep = 0.0
                    lsiter = 0
                    maxlsiter = self.maxlsiter
                    updated = False
                    minimum = False
                    failed = True
                    lsmethod = 'CC'
            else:
                maxlsiter = 0
                lsiter = -1
                optstep = epsstep
                updated = False
                minimum = True
                failed = False
                lsmethod = 'CC-EPS'

            if (optstep == 0.0):
                #
                # we are in the concave region or force-only estimate failed,
                # just follow the (conjugate) gradient
                step = dltstep * abs(step)
                U_mm = matrix_exponential(D_mm, step)
                self.W_mn = np.dot(U_mm, W_old_mn)
                self.update_optimal_states()
                self.update_potentials()
                E1 = self.esic
                #
                #
                if (abs(E1 - E0) < prec and E1 >= E0):
                    ESI = E1
                    optstep = 0.0
                    updated = False
                    minimum = True
                    failed = True
                    lsmethod = lsmethod + '-DLT-N'
                    noise = True
                    maxlsiter = -1
                elif (E1 < E0):
                    ESI = E1
                    optstep = step
                    updated = True
                    minimum = False
                    failed = False
                    lsmethod = lsmethod + '-DLT'
                    maxlsiter = self.maxlsiter
                elif (eps_works):
                    ESI = Eeps
                    E1 = Eeps
                    step = epsstep
                    updated = False
                    minimum = False
                    failed = False
                    lsmethod = lsmethod + '-EPS'
                    maxlsiter = self.maxlsiter
                else:
                    optstep = 0.0
                    updated = False
                    minimum = False
                    failed = True
                    lsmethod = lsmethod + '-EPS-F'
                    maxlsiter = -1
                #
                G = (E1 - E0) / step
                step0 = 0.0
                step1 = step
                step2 = 2 * step
                #
                for lsiter in range(maxlsiter):
                    #
                    # energy at the new position
                    U_mm = matrix_exponential(D_mm, step2)
                    self.W_mn = np.dot(U_mm, W_old_mn)
                    self.update_optimal_states()
                    self.update_potentials()
                    E2 = self.esic
                    G = (E2 - E1) / (step2 - step1)
                    #
                    #
                    if (G > 0.0):
                        if self.lsinterp:
                            a = (E0 / ((step2 - step0) * (step1 - step0)) +
                                 E2 / ((step2 - step1) * (step2 - step0)) -
                                 E1 / ((step2 - step1) * (step1 - step0)))
                            b = (E2 - E0) / (step2 - step0) - a * (step2 +
                                                                   step0)
                            if (a < step**2):
                                optstep = 0.5 * (step0 + step2)
                            else:
                                optstep = -0.5 * b / a
                            updated = False
                            minimum = True
                            break
                        else:
                            optstep = step1
                            updated = False
                            minimum = True
                            break

                    elif (G < 0.0):
                        optstep = step2
                        step0 = step1
                        step1 = step2
                        step2 = step2 + step
                        E0 = E1
                        E1 = E2
                        ESI = E2
                        updated = True
                        minimum = False

            if (cgiter != 0):
                lsmethod = lsmethod + ' CG'

            if (cgiter >= self.maxcgiter or not minimum):
                Gold = 0.0
                cgiter = 0
            else:
                cgiter = cgiter + 1
                D_old_mm[:, :] = D_mm[:, :]

            # update the energy and matrixelements of V and Kappa
            # and accumulate total residual of unitary optimization
            if (not updated):
                if optstep > 0.0:
                    U_mm = matrix_exponential(D_mm, optstep)
                    self.W_mn = np.dot(U_mm, W_old_mn)
                    self.update_optimal_states()
                    self.update_potentials()
                    ESI = self.esic
                    V_mm, K_mm, norm = self.calculate_sic_matrixelements()
                else:
                    self.W_mn = W_old_mn
                    self.update_optimal_states()
                    self.update_potentials(restore=True)
                    ESI = self.esic
                    V_mm, K_mm, norm = self.calculate_sic_matrixelements()

            if (lsiter == maxlsiter - 1):
                V_mm, K_mm, norm = self.calculate_sic_matrixelements()

            E0 = ESI

            # orthonormalize the energy optimal orbitals
            self.W_mn = ortho(self.W_mn)
            K = max(norm, 1.0e-16)

            if self.finegd.comm.rank == 0:
                if self.logging == 1:
                    print("           UO-%d: %2d %5.1f  %20.5f  " %
                          (self.spin, iter, np.log10(K), ESI * Hartree))
                elif self.logging == 2:
                    print("           UO-%d: %2d %5.1f  %20.5f  " %
                          (self.spin, iter, np.log10(K), ESI * Hartree) +
                          lsmethod)

            if ((K < self.uomaxres or
                 K < self.wfs.eigensolver.error * self.uorelres) or noise or
                    failed) and not self.maxuoiter < 0:
                break



import pytest
from ase.build import molecule

from gpaw import GPAW, FermiDirac, PoissonSolver
from gpaw.test import equal, gen


@pytest.mark.later
def test_corehole_h2o_dks():
    def xc(name):
        return {'name': name, 'stencil': 1}

    # Generate setup for oxygen with a core-hole:
    s = gen('O', name='fch1s', xcname='PBE', corehole=(1, 0, 1.0))

    atoms = molecule('H2O')
    atoms.center(vacuum=2.5)

    params = dict(mode='fd',
                  xc=xc('PBE'),
                  poissonsolver=PoissonSolver('fd',
                                              use_charge_center=True))
    atoms.calc = GPAW(**params)
    e1 = atoms.get_potential_energy() + atoms.calc.get_reference_energy()

    atoms[0].magmom = 1
    atoms.calc = GPAW(charge=-1,
                      setups={'O': s},
                      occupations=FermiDirac(0.0, fixmagmom=True),
                      **params)
    e2 = atoms.get_potential_energy() + atoms.calc.get_reference_energy()

    atoms[0].magmom = 0
    atoms.calc = GPAW(charge=0,
                      setups={'O': s},
                      occupations=FermiDirac(0.0, fixmagmom=True),
                      spinpol=True,
                      **params)
    e3 = atoms.get_potential_energy() + atoms.calc.get_reference_energy()

    print('Energy difference %.3f eV' % (e2 - e1))
    print('XPS %.3f eV' % (e3 - e1))

    print(e2 - e1)
    print(e3 - e1)
    assert abs(e2 - e1 - 533.070) < 0.001
    assert abs(e3 - e1 - 538.549) < 0.001

    energy_tolerance = 0.02
    print(e1)
    print(e2)
    print(e3)
    equal(e1, -2080.3651, energy_tolerance)
    equal(e2, -1547.2944, energy_tolerance)
    equal(e3, -1541.8152, energy_tolerance)



import pytest
from ase import Atom, Atoms
from ase.parallel import world

from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT


def get_H2(calculator=None):
    """Define H2 and set calculator if given"""
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 4.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    if calculator is not None:
        H2.calc = calculator

    return H2


def run_and_delete(txt):
    outfname = 'gpawlog.txt'
    calc = GPAW(mode='fd', xc='PBE', h=0.25, nbands=5, txt=outfname)
    calc.calculate(get_H2(calc))
    exlst = LrTDDFT(calc, restrict={'eps': 0.4, 'jend': 3}, txt=txt)
    del calc
    del exlst
    world.barrier()


@pytest.mark.lrtddft
def test_log(in_tmp_dir):
    defname = 'gpawlog.txt'
    # LrTDDFT outputs to the same log like gpaw
    run_and_delete(txt=defname)
    if world.rank == 0:
        with open(defname) as f:
            string = f.read()
            assert 'Kohn-Sham single transitions' in string
            assert 'Linear response TDDFT calculation' in string
    world.barrier()

    # silent LrTDDFT
    run_and_delete(txt=None)
    if world.rank == 0:
        with open(defname) as f:
            assert 'RPA kss[0]' not in f.read()
    world.barrier()

    # output to own file
    ownfname = 'lrtddftlog.txt'
    run_and_delete(txt=ownfname)

    if world.rank == 0:
        with open(defname) as f:
            assert 'Linear response TDDFT calculation' not in f.read()
        with open(ownfname) as f:
            string = f.read()
            assert 'Kohn-Sham single transitions' in string
            assert 'Linear response TDDFT calculation' in string
    world.barrier()



from __future__ import annotations

from collections import defaultdict
from typing import Sequence

import numpy as np
from ase.data import covalent_radii
from ase.data.colors import jmol_colors
from ase.units import Bohr, Hartree
from gpaw.calculator import GPAW
from gpaw.lcao.tightbinding import TightBinding  # as LCAOTightBinding
from gpaw.lcao.tools import get_bfi
from gpaw.typing import Array1D, Array2D, Array4D
from gpaw.utilities.blas import r2k
from gpaw.utilities.tools import lowdin, tri2full
from scipy.linalg import eigh


def get_subspace(A_MM: Array2D, indices: Sequence[int]):
    """Get the subspace spanned by the basis function listed in index."""
    assert A_MM.ndim == 2 and A_MM.shape[0] == A_MM.shape[1]
    return A_MM.take(indices, 0).take(indices, 1)


def get_orthonormal_subspace(H_MM: Array2D,
                             S_MM: Array2D,
                             indices: Sequence[int] = None):
    """Get orthonormal eigenvalues and -vectors of subspace listed in index."""
    if indices is not None:
        h_ww = get_subspace(H_MM, indices)
        s_ww = get_subspace(S_MM, indices)
    else:
        h_ww = H_MM
        s_ww = S_MM
    eps, v = eigh(h_ww, s_ww)
    return eps, v


def subdiagonalize(H_MM: Array2D,
                   S_MM: Array2D,
                   blocks: Sequence[Sequence[int]]):
    """Subdiagonalize blocks."""
    nM = len(H_MM)
    v_MM = np.eye(nM)
    eps_M = np.zeros(nM)
    mask_M = np.ones(nM, dtype=int)
    for block in blocks:
        eps, v = get_orthonormal_subspace(H_MM, S_MM, block)
        v_MM[np.ix_(block, block)] = v
        eps_M[block] = eps
        mask_M[block] = 0
    epsx_M = np.ma.masked_array(eps_M, mask=mask_M)  # type: ignore
    return epsx_M, v_MM


def subdiagonalize_atoms(calc: GPAW,
                         H_MM: Array2D,
                         S_MM: Array2D,
                         atom_indices: int | Sequence[int] | None = None):
    """Subdiagonalize atomic sub-spaces."""
    if atom_indices is None:
        atom_indices = range(len(calc.atoms))
    if isinstance(atom_indices, int):
        atom_indices = [atom_indices]
    block_lists = []
    for a in atom_indices:
        M = calc.wfs.basis_functions.M_a[a]
        block = range(M, M + calc.wfs.setups[a].nao)
        block_lists.append(block)
    return subdiagonalize(H_MM, S_MM, block_lists)


def get_orbitals(calc: GPAW, U_Mw: Array2D, q: int = 0):
    """Get orbitals from AOs coefficients.

    Parameters
    ----------
    calc : GPAW
        LCAO calculator
    U_Mw : array_like
        LCAO expansion coefficients.
    """
    Nw = U_Mw.shape[1]
    C_wM = np.ascontiguousarray(U_Mw.T).astype(calc.wfs.dtype)
    w_wG = calc.wfs.gd.zeros(Nw, dtype=calc.wfs.dtype)
    calc.wfs.basis_functions.lcao_to_grid(C_wM, w_wG, q=q)
    return w_wG


def get_xc(calc: GPAW, v_wG: Array4D, P_awi=None):
    """Get exchange-correlation part of the Hamiltonian."""
    if calc.density.nt_sg is None:
        calc.density.interpolate_pseudo_density()
    nt_sg = calc.density.nt_sg
    vxct_sg = calc.density.finegd.zeros(calc.wfs.nspins)
    calc.hamiltonian.xc.calculate(calc.density.finegd, nt_sg, vxct_sg)
    vxct_G = calc.wfs.gd.empty()
    calc.hamiltonian.restrict_and_collect(vxct_sg[0], vxct_G)

    # Integrate pseudo part
    Nw = len(v_wG)
    xc_ww = np.empty((Nw, Nw))
    r2k(0.5 * calc.wfs.gd.dv, v_wG, vxct_G * v_wG, 0.0, xc_ww)
    tri2full(xc_ww, "L")

    # Atomic PAW corrections required? XXX
    if P_awi is not None:
        raise NotImplementedError(
            'Atomic PAW corrections not included. '
            'Have a look at pwf2::get_xc2 for inspiration.')

    return xc_ww * Hartree


def get_Fcore():
    pass


class BasisTransform:
    """Class to perform a basis transformation.

    Attributes
    ----------
    U_MM : array_like
        2-D rotation matrix between 2 basis.
    indices : array_like, optional
        1-D array of sub-indices of the new basis.
    U_Mw : array_like, optional
        Same as `U_MM` but includes only `indices` of the new basis.

    Methods
    -------
    rotate_matrx(A_MM, keep_rest=False)
        Rotate a matrix.
    rotate_projections(P_aMi, keep_rest=False)
        Rotate PAW atomic projects.
    rotate_function(P_aMi, keep_rest=False)
        Rotate PAW atomic projects.

    """

    def __init__(self, U_MM: Array2D, indices: Sequence[int] = None) -> None:
        """

        Parameters
        ----------
        See class docstring

        """
        self.U_MM = U_MM
        self.indices = indices
        self.U_Mw: Array2D | None
        if indices is not None:
            self.U_Mw = np.ascontiguousarray(U_MM[:, indices])
        else:
            self.U_Mw = None

    def get_rotation(self, keep_rest: bool = False):
        if keep_rest or self.U_Mw is None:
            return self.U_MM
        return self.U_Mw

    def rotate_matrix(self, A_MM: Array2D, keep_rest: bool = False):
        U_Mx = self.get_rotation(keep_rest)
        return U_Mx.T.conj() @ A_MM @ U_Mx

    def rotate_projections(self, P_aMi, keep_rest: bool = False):
        U_Mx = self.get_rotation(keep_rest)
        P_awi = {}
        for a, P_Mi in P_aMi.items():
            P_awi[a] = np.tensordot(U_Mx, P_Mi, axes=([0], [0]))
        return P_awi

    def rotate_function(self, Psi_MG: Array4D, keep_rest: bool = False):
        U_Mx = self.get_rotation(keep_rest)
        return np.tensordot(U_Mx, Psi_MG, axes=([0], [0]))


class EffectiveModel(BasisTransform):
    """Class for an effective model.

    See Also
    --------
    BasisTranform


    Methods
    -------
    get_static_correction(H_MM: npt.NDArray,
                          S_MM: npt.NDArray,
                          z: complex = 0. + 1e-5j)
        Hybridization of the effective model with the rest evaluated at `z`.

    """

    def __init__(self,
                 U_MM: Array2D,
                 indices: Sequence[int],
                 S_MM: Array2D = None) -> None:
        """

        See Also
        --------
        BasisTransform

        Parameters
        ----------
        S_MM : array_like, optional
            2-D LCAO overlap matrix. If provided, the resulting basis
            is orthogonalized.
        """
        if S_MM is not None:
            lowdin(self.U_Mw, self.rotate_matrix(S_MM))
            assert self.U_Mw is not None
            np.testing.assert_allclose(self.rotate_matrix(
                S_MM), np.eye(self.U_Mw.shape[1]))
            U_MM = U_MM[:]
            U_MM = U_MM[:, indices] = self.U_Mw

        super().__init__(U_MM, indices)

    def get_static_correction(self,
                              H_MM: Array2D,
                              S_MM: Array2D,
                              z: complex = 0. + 1e-5j):
        """Get static correction to model Hamiltonian.

        Parameters
        ----------
        H_MM, S_MM : array_like
            2-D LCAO Hamiltonian and overlap matrices.
        z : complex
            Energy with a small positive immaginary shift.

        """
        w = self.indices  # Alias
        assert w is not None

        Hp_MM = self.rotate_matrix(H_MM, keep_rest=True)
        Sp_MM = self.rotate_matrix(S_MM, keep_rest=True)
        Up_Mw = Sp_MM[:, w].dot(np.linalg.inv(Sp_MM[np.ix_(w, w)]))

        H_ww = self.rotate_matrix(H_MM)
        S_ww = self.rotate_matrix(S_MM)

        # Coupled
        G = np.linalg.inv(z * Sp_MM - Hp_MM)
        # G_inv = np.linalg.inv(rotate_matrix(G, Up_Mw))
        G_inv = np.linalg.inv(Up_Mw.T.conj() @ G @ Up_Mw)
        # Uncoupled
        G0_inv = z * S_ww - H_ww
        # Hybridization
        D0 = G0_inv - G_inv
        return D0.real

    def __len__(self):
        return len(self.indices)


class Subdiagonalization(BasisTransform):
    """Class to perform a subdiagonalization of the Hamiltonian.

    Attributes
    ----------
    blocks : list of array_like
        List of blocks to subdiagonalize.
    H_MM, S_MM : array_like
        2-D LCAO Hamiltonian and overlap matrices.
    U_MM : array_like
        2-D rotation matrix that subdiagonalizes the LCAO Hamiltonian.
    eps_M : array_like
        1-D array of local orbital energies.

    Methods
    -------
    group_energies(round=1)
        Group local orbitals based on energy
    group_symmetries(cutoff=0.9)
        Group local orbitals based on symmetries and energy.
    get_effective_model(indices, ortho=None)
        Builds and effective model from an array of indices.

    """

    def __init__(self,
                 H_MM: Array2D,
                 S_MM: Array2D,
                 blocks: Sequence[Sequence[int]]) -> None:
        """

        Parameters
        ----------
        See class docstring

        """
        self.blocks = blocks
        self.H_MM = H_MM
        self.S_MM = S_MM
        self.eps_M, U_MM = subdiagonalize(
            self.H_MM, self.S_MM, blocks)
        super().__init__(U_MM)
        # Groups of local orbitals with the same symmetry.
        self.groups: dict[float, list[int]] | None = None

    def group_energies(self, decimals: int = 1):
        """Group local orbitals with a similar energy.

        Parameters
        ----------
        decimals : int
            Round energies to the given number of decimals.

        """
        eps = self.eps_M.round(decimals)
        show = np.where(~eps.mask)[0]
        groups = defaultdict(list)
        for index in show:
            groups[eps[index]].append(index)

        self.groups = groups
        return self.groups

    def group_symmetries(self, decimals: int = 1, cutoff: float = 0.9):
        """Group local orbitals with a similar spatial symmetry and energy.

        Parameters
        ----------
        decimals : int
            Round energies to the given number of decimals.
        cutoff : float
            Sets minimum degree of overlap. Can be any value between 0 and 1.

        """
        col_1: list[int] = []  # Keyword.
        col_2: list[int] = []  # Value.
        groups = defaultdict(set)
        blocks = self.blocks
        # Loop over pair of blocks.
        for b1, b2 in zip(*np.triu_indices(len(blocks), k=1)):
            if len(blocks[b1]) != len(blocks[b2]):
                # Blocks with different dimensions not compatible.
                continue
            U1 = self.U_MM[np.ix_(blocks[b1], blocks[b1])]
            U2 = self.U_MM[np.ix_(blocks[b2], blocks[b2])]
            # Compute pair overlap between orbitals in the two blocks.
            for o1, o2 in np.ndindex(len(blocks[b1]), len(blocks[b1])):
                v1 = abs(U1[:, o1])
                v2 = abs(U2[:, o2])
                o12 = 2 * v1.dot(v2) / (v1.dot(v1) + v2.dot(v2))
                # Overlap larger than cutoff?
                if o12 >= cutoff:
                    # Yes.
                    i1 = blocks[b1][o1]
                    i2 = blocks[b2][o2]
                    # Use orbital with minimal index as keyword.
                    i1, i2 = min(i1, i2), max(i1, i2)
                    # Check if `i1` is already present in `col_2` and
                    # use corresponding keyword in col_1 instead.
                    present = False
                    for i, i3 in enumerate(col_2):
                        if i1 == i3:
                            present = True
                            break
                    if present:
                        a1 = col_1[i]
                    else:
                        a1 = i1
                    col_1.append(a1)
                    col_2.append(i2)
                    groups[a1].add(i2)
        # Try to further group by energy.
        new: dict[float, list[int]] = defaultdict(list)
        for k, v in groups.items():
            v.add(k)
            new[self.eps_M[k].round(decimals)] += groups[k]
        self.groups = {k: list(sorted(new[k])) for k in sorted(new)}  # groups
        return self.groups

    def get_model(self,
                  indices: Sequence[int],
                  ortho: bool = False) -> EffectiveModel:
        """Extract an effective model from the subdiagonalized space.

        Parameters
        ----------
        indices : array_like
            1-D array of indices to include in the model from
            the new basis.
        ortho : bool, default=False
            Whether to orthogonalize the model basis.
        """
        return EffectiveModel(self.U_MM, indices, self.S_MM if ortho else None)


class LocalOrbitals(TightBinding):
    """Local Orbitals.

    Attributes
    ----------
    TODO

    Methods:
    --------
    subdiagonalize(self, symbols=None, blocks=None, groupby='energy')
        Subdiagonalize the LCAO Hamiltonian.
    take_model(self, indices=None, minimal=True, cutoff=1e-3, ortho=False)
        Take an effective model of local orbitals.
    TODO

    """

    def __init__(self, calc: GPAW):
        self.calc = calc
        self.gamma = calc.wfs.kd.gamma  # Gamma point calculation
        self.subdiag: Subdiagonalization | None = None
        self.model: EffectiveModel | None = None

        if self.gamma:
            self.calc = calc
            h = self.calc.hamiltonian
            wfs = self.calc.wfs
            kpt = wfs.kpt_u[0]

            H_MM = wfs.eigensolver.calculate_hamiltonian_matrix(h, wfs, kpt)
            S_MM = wfs.S_qMM[kpt.q]
            # XXX Converting to full matrices here
            tri2full(H_MM)
            tri2full(S_MM)
            self.H_NMM = H_MM[None, ...] * Hartree  # eV
            self.S_NMM = S_MM[None, ...]
            self.N0 = 0
        else:
            super().__init__(calc.atoms, calc)
            # Bloch to real
            self.H_NMM, self.S_NMM = TightBinding.h_and_s(self)
            self.H_NMM *= Hartree  # eV
            try:
                self.N0 = int(np.argwhere(
                    self.R_cN.T.dot(self.R_cN) < 1e-13).flat[0])
            except Exception as exc:
                raise RuntimeError(
                    "Must include central unit cell, i.e. R=[0,0,0].") from exc

    def subdiagonalize(self,
                       symbols: Array1D = None,
                       blocks: Sequence[Sequence[int]] = None,
                       groupby: str = 'energy'):
        """Subdiagonalize Hamiltonian and overlap matrices.

        Parameters
        ----------
        symbols : array_like, optional
            Element or elements to subdiagonalize.
        blocks : list of array_like, optional
            List of blocks to subdiagonalize.
        groupby : {'energy,'symmetry'}, optional
            Group local orbitals based on energy or
            symmetry and energy. Default is 'energy'.

        """
        if symbols is not None:
            atoms = self.calc.atoms.symbols.search(symbols)
            blocks = [get_bfi(self.calc, [c]) for c in atoms]
        if blocks is None:
            raise RuntimeError("""User must provide either the element(s)
                               or a list of blocks to subdiagonalize.""")
        self.blocks = blocks
        self.subdiag = Subdiagonalization(
            self.H_NMM[self.N0], self.S_NMM[self.N0], blocks)

        self.groupby(groupby)

    def groupby(self,
                method: str = 'energy',
                decimals: int = 1,
                cutoff: float = 0.9):
        """Group local orbitals by symmetry.

        Parameters
        ----------
        method : {'energy,'symmetry'}, optional
            Group local orbitals based on energy or
            symmetry and energy. Default is 'energy'.
        decimals, cutoff : optional
            Parameters passed to the group methods.

        """
        assert self.subdiag is not None
        if method == 'energy':
            self.groups = self.subdiag.group_energies(decimals=decimals)
        elif method == 'symmetry':
            self.groups = self.subdiag.group_symmetries(
                decimals=decimals, cutoff=cutoff)
        else:
            raise RuntimeError(
                f"Invalid method type. {method} not in {'energy', 'symmetry'}")
        # Ensure previous model is invalid.
        self.model = None

    def take_model(self,
                   indices: list[int] = None,
                   minimal: bool = True,
                   cutoff: float = 1e-3,
                   ortho: bool = False):
        """Build an effective model.

        Parameters
        ----------
        indices : array_like
            1-D array of indices to include in the model
            from the new basis.
        minimal : bool, default=True
            Whether to add (minimal=False) or not (minimal=True)
            the orbitals with an overlap larger than `cuoff` with any of the
            orbital specified by `indices`.
        cutoff : float
            Cutoff value for the maximum matrix element connecting a group
            with the minimal model.
        ortho : bool, default=False
            Whether to orthogonalize the model.

        """
        if self.subdiag is None:
            raise RuntimeError("""Not yet subdiagonalized.""")

        eps = self.subdiag.eps_M.round(1)
        indices_from_input = indices is not None

        if indices is None:
            # Find active orbitals with energy closest to Fermi.

            fermi = round(self.calc.get_fermi_level(), 1)
            # diffs = [] # Min distance from fermi for each block
            indices = []  # Min distance index for each block
            for block in self.blocks:
                eb = eps[block]
                ib = np.abs(eb - fermi).argmin()
                indices.append(block[ib])
                # diffs.append(abs(eb[ib]))

        if not minimal:
            # Find orbitals that connect to active with a matrix
            # element larger than cutoff

            # Look at gamma of 1st neighbor
            H_MM = self.H_NMM[(self.N0 + 1) % len(self.H_NMM)]
            H_MM = self.subdiag.rotate_matrix(H_MM)
            # H_MM = dots(self.subdiag.U_MM.T.conj(), H_MM, self.subdiag.U_MM)

            extend = []
            for group in self.groups.values():
                if np.isin(group, indices).any():
                    continue
                if np.abs(H_MM[np.ix_(indices, group)]).max() > cutoff:
                    extend += group

            # Expand model
            indices += extend

        self.indices = indices
        self.model = self.subdiag.get_model(indices, ortho=ortho)

        if self.gamma:
            H_Nww = self.model.rotate_matrix(self.H_NMM[0])[None, ...]
            S_Nww = self.model.rotate_matrix(self.S_NMM[0])[None, ...]

        else:
            # Bypass parent's LCAO construction.
            shape = (self.R_cN.shape[1],) + 2 * (len(self.indices),)
            dtype = self.H_NMM.dtype
            H_Nww = np.empty(shape, dtype)
            S_Nww = np.empty(shape, dtype)

            for N, (H_MM, S_MM) in enumerate(zip(self.H_NMM, self.S_NMM)):
                H_Nww[N] = self.model.rotate_matrix(H_MM)
                S_Nww[N] = self.model.rotate_matrix(S_MM)
        self.H_Nww = H_Nww
        self.S_Nww = S_Nww

        if minimal and not indices_from_input:
            print("Add static correction.")
            # Add static correction of hybridization to minimal model.
            self.H_Nww[self.N0] += self.model.get_static_correction(
                self.H_NMM[self.N0], self.S_NMM[self.N0])

    def h_and_s(self):
        # Hartree units.
        # Bypass TightBinding method.
        eV = 1 / Hartree
        return self.H_Nww * eV, self.S_Nww

    def band_structure(self, path_kc, blochstates=False):
        # Broute force hack to restore matrices.
        H_NMM = self.H_NMM
        S_NMM = self.S_NMM
        ret = TightBinding.band_structure(self, path_kc, blochstates)
        self.H_NMM = H_NMM
        self.S_NMM = S_NMM
        return ret

    def get_hamiltonian(self):
        """Get the Hamiltonian in the home unit cell."""
        return self.H_Nww[self.N0]

    def get_overlap(self):
        """Get the overlap in the home unit cell."""
        return self.S_Nww[self.N0]

    def get_orbitals(self, indices):
        """Get orbitals on the real-space grid."""
        if self.model is None:
            basis = self.subdiag
        else:
            # Maybe model is orthogonal and subdiag does not know.
            basis = self.model
        return get_orbitals(self.calc, basis.U_MM[:, indices])

    def plot_group(self, group):
        return plot2D_orbitals(self, self.groups[group])

    def get_projections(self, q=0):
        P_aMi = {a: P_aqMi[q] for a, P_aqMi in self.calc.wfs.P_aqMi.items()}
        return self.model.rotate_projections(P_aMi)

    def get_xc(self):
        return get_xc(self.calc, self.get_orbitals(), self.get_projections())

    def get_Fcore(self):
        pass


# Plotting tools


def get_plane_dirs(plane):
    """Get normal and in-plane directions for a plane identified
    by any combination of {'x','y','z'}.

    Parameters
    ----------
    plane : str
        Pair of chars identifying the plane, e.g. 'xy'.

    Returns
    -------
    norm_dir : int
        Normal direction
    norm_dir : list
        In-plane directions
    """
    plane_dirs = ['xyz'.index(i) for i in sorted(plane)]
    norm_dir = [i for i in [0, 1, 2] if i not in plane_dirs]
    return norm_dir[0], plane_dirs


def get_atoms(calc, indices):
    """Get the list of atoms corresponding to the given indices.

    Parameters
    ----------
    calc : GPAW
        Calculator
    indices : array_like
        List of orbitals for which to retreive the atoms to which they belong.

    Returns
    -------
    a_list : list
        List of atoms for each index.
    unique : list
        Indices of first occurrences of an atom.
    """
    atoms = calc.atoms
    a_list = []
    unique = []
    for a in range(len(atoms)):
        matches = np.where(np.isin(get_bfi(calc, [a]), indices))[0]
        if len(matches) > 0:
            a_list += [a] * len(matches)
            unique.append(len(a_list) - len(matches))
    return a_list, unique


def plot2D_orbitals(los, indices, plane='yz'):
    """Plot a 2D slice of the orbitals.

    Parameters
    ----------
    los : LocalOrbitals
        Local orbital wrapper
    indices : array_like
        List of orbitals to display
    plane : str, optional
        Pair of chars identifying the plane, by default 'yz'.

    Returns
    -------
    _type_
        _description_
    """

    import matplotlib.pyplot as plt

    norm_dir, plane_dirs = get_plane_dirs(plane)

    calc = los.calc
    atoms = los.calc.atoms

    def get_coord(c):
        return calc.wfs.gd.coords(c, pad=False) * Bohr

    w_wG = los.get_orbitals(indices)

    radii = covalent_radii[calc.atoms.numbers]
    colors = jmol_colors[calc.atoms.numbers]
    pos = atoms.positions

    # Take planes at atomic positions.
    a_list, _ = get_atoms(calc, indices)
    slice_planes = np.searchsorted(get_coord(norm_dir), pos[a_list, norm_dir])

    # Take box limited by external atoms plus 4 Ang vacuum.
    box_lims = [(l - 2, u + 2) for l, u in zip(pos[:, plane_dirs].min(0),
                                               pos[:, plane_dirs].max(0))]
    box_widths = [lims[1] - lims[0] for lims in box_lims]
    ratio = box_widths[0] / box_widths[1]  # Cell ratio
    num_orbs = len(indices)
    max_cols = 6
    nrows = (num_orbs - 1) // max_cols + 1
    ncols = min(num_orbs, max_cols)

    figsize = 5
    fig, axs = plt.subplots(nrows, ncols, figsize=(
        ncols / nrows * ratio * figsize, figsize))

    X, Y = np.meshgrid(get_coord(plane_dirs[0]), get_coord(
        plane_dirs[1]), indexing='ij')
    take_plane = [slice(None)] * 3

    it = np.nditer(axs, flags=['refs_ok', 'c_index', 'multi_index'])
    for _ in it:
        ax = axs[it.multi_index]
        w = np.ravel_multi_index(it.multi_index, axs.shape)
        if w >= num_orbs:
            ax.axis('off')
            continue

        take_plane[norm_dir] = slice_planes[w]
        C = w_wG[w][tuple(take_plane)]

        ax.pcolormesh(X, Y, C, cmap='jet', shading='gouraud')
        ax.set_xlim(box_lims[0])
        ax.set_ylim(box_lims[1])
        ax.axis('off')

        ax.scatter(pos[:, plane_dirs[0]], pos[:, plane_dirs[1]],
                   c=colors, s=radii * 1e3 / (nrows * 2))
    return fig



from ase import Atoms
from gpaw import GPAW
from gpaw.mpi import size


def test_generic_al_chain(in_tmp_dir):
    d = 4.0 / 2**0.5
    ndomains = size // 8 + 1
    calc = GPAW(mode='fd',
                h=d / 16,
                kpts=(17, 1, 1),
                parallel={'domain': ndomains, 'band': 1})
    chain = Atoms('Al', cell=(d, 5, 5), pbc=True, calculator=calc)
    e = chain.get_potential_energy()
    print(e)
    assert abs(e - -1.8182) < 0.0005
    assert calc.wfs.kd.comm.size * ndomains == size
    calc.write('al_chain', 'all')



import pytest
from ase import Atoms
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.test import equal
from gpaw.xc.hybrid import HybridXC


@pytest.mark.libxc
@pytest.mark.hybrids
def test_exx_unocc():

    loa = Atoms('Be2',
                [(0, 0, 0), (2.45, 0, 0)],
                cell=[5.9, 4.8, 5.0])
    loa.center()

    txt = None
    xc = 'PBE0'
    nbands = 4

    unocc = True
    load = False

    base_params = dict(
        mode='fd',
        h=0.3,
        eigensolver='rmm-diis',
        nbands=nbands,
        convergence={'eigenstates': 1e-4},
        txt=txt)
    # usual calculation
    fname = 'Be2.gpw'
    if not load:
        xco = HybridXC(xc)
        cocc = GPAW(**base_params, xc=xco)
        cocc.calculate(loa)
    else:
        cocc = GPAW(fname)
        cocc.converge_wave_functions()
    fo_n = 1. * cocc.get_occupation_numbers()
    eo_n = 1. * cocc.get_eigenvalues()

    if unocc:
        # apply Fock opeartor also to unoccupied orbitals
        xcu = HybridXC(xc, unocc=True)
        cunocc = GPAW(**base_params, xc=xcu)
        cunocc.calculate(loa)

        parprint('     HF occ          HF unocc      diff')
        parprint('Energy %10.4f   %10.4f %10.4f' %
                 (cocc.get_potential_energy(),
                  cunocc.get_potential_energy(),
                  cocc.get_potential_energy() - cunocc.get_potential_energy()
                  ))
        equal(cocc.get_potential_energy(),
              cunocc.get_potential_energy(), 1.e-4)

        fu_n = cunocc.get_occupation_numbers()
        eu_n = cunocc.get_eigenvalues()

        parprint('Eigenvalues:')
        for eo, fo, eu, fu in zip(eo_n, fo_n, eu_n, fu_n):
            parprint('%8.4f %5.2f   %8.4f %5.2f  %8.4f' %
                     (eo, fo, eu, fu, eu - eo))
            if fo > 0.01:
                equal(eo, eu, 3.5e-4)



import numpy as np
from gpaw.lfc import LocalizedFunctionsCollection as LFC
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline


def test_lfc_derivatives():
    a = 4.0
    gd = GridDescriptor(N_c=[16, 20, 20], cell_cv=[a, a + 1, a + 2],
                        pbc_c=(0, 1, 1))
    spos_ac = np.array([[0.25, 0.15, 0.35], [0.5, 0.5, 0.5]])
    s = Spline(l=0, rmax=2.0, f_g=np.array([1, 0.9, 0.1, 0.0]))
    p = Spline(l=1, rmax=2.0, f_g=np.array([1, 0.9, 0.1, 0.0]))
    spline_aj = [[s], [s, p]]
    c = LFC(gd, spline_aj, cut=True, forces=True)
    c.set_positions(spos_ac)
    C_ani = c.dict(3, zero=True)
    if 1 in C_ani:
        C_ani[1][:, 1:] = np.eye(3)
    psi = gd.zeros(3)
    c.add(psi, C_ani)
    c.integrate(psi, C_ani)
    if 1 in C_ani:
        d = C_ani[1][:, 1:].diagonal()
        assert d.ptp() < 4e-6
        C_ani[1][:, 1:] -= np.diag(d)
        assert abs(C_ani[1]).max() < 5e-17
    d_aniv = c.dict(3, derivative=True)
    c.derivative(psi, d_aniv)
    if 1 in d_aniv:
        for v in range(3):
            assert abs(d_aniv[1][v - 1, 0, v] + 0.2144) < 5e-5
            d_aniv[1][v - 1, 0, v] = 0
        assert abs(d_aniv[1]).max() < 3e-16
    eps = 0.0001
    pos_av = np.dot(spos_ac, gd.cell_cv)
    for v in range(3):
        pos_av[0, v] += eps
        c.set_positions(np.dot(pos_av, gd.icell_cv.T))
        c.integrate(psi, C_ani)
        if 0 in d_aniv:
            C0_n = C_ani[0][:, 0].copy()
        pos_av[0, v] -= 2 * eps
        c.set_positions(np.dot(pos_av, gd.icell_cv.T))
        c.integrate(psi, C_ani)
        if 0 in d_aniv:
            C0_n -= C_ani[0][:, 0]
            C0_n /= -2 * eps
            assert abs(C0_n - d_aniv[0][:, 0, v]).max() < 1e-8
        pos_av[0, v] += eps
